# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import logging
from abc import ABC, abstractmethod
from pathlib import Path
from time import monotonic
from typing import Iterable, cast

from datacube.model import MetadataType
from datacube.utils import read_documents, jsonify_document, InvalidDocException
from datacube.utils.changes import DocumentMismatchError, check_doc_unchanged, Change
from datacube.utils.documents import JsonDict

from ._types import BatchStatus

_LOG = logging.getLogger(__name__)

_DEFAULT_METADATA_TYPES_PATH = Path(__file__).parent.joinpath('default-metadata-types.yaml')


def default_metadata_type_docs(path=_DEFAULT_METADATA_TYPES_PATH) -> list[MetadataType]:
    """A list of the bare dictionary format of default :class:`datacube.model.MetadataType`"""
    return [doc for (path, doc) in read_documents(path)]


class AbstractMetadataTypeResource(ABC):
    """
    Abstract base class for the MetadataType portion of an index api.

    All MetadataTypeResource implementations should inherit from this base
    class and implement all abstract methods.

    (If a particular abstract method is not applicable for a particular implementation
    raise a NotImplementedError)
    """

    @abstractmethod
    def from_doc(self, definition: JsonDict) -> MetadataType:
        """
        Construct a MetadataType object from a dictionary definition

        :param definition: A metadata definition dictionary
        :return: An unpersisted MetadataType object
        """

    @abstractmethod
    def add(self,
            metadata_type: MetadataType,
            allow_table_lock: bool = False
           ) -> MetadataType:
        """
        Add a metadata type to the index.

        :param metadata_type: Unpersisted Metadatatype model
        :param allow_table_lock:
            Allow an exclusive lock to be taken on the table while creating the indexes.
            This will halt other user's requests until completed.

            If false, creation will be slightly slower and cannot be done in a transaction.

            raise NotImplementedError if set to True, and this behaviour is not applicable
            for the implementing driver.
        :return: Persisted Metadatatype model.
        """

    def _add_batch(self, batch_types: Iterable[MetadataType]) -> BatchStatus:
        """
        Add a single "batch" of mdts.

        Default implementation is simple loop of add

        API Note: This API method is not finalised and may be subject to change.

        :param batch_types: An iterable of one batch's worth of MetadataType objects to add
        :return: BatchStatus named tuple.
        """
        b_skipped = 0
        b_added = 0
        b_started = monotonic()
        b_loaded = set()
        for mdt in batch_types:
            try:
                self.add(mdt)
                b_added += 1
                b_loaded.add(mdt.name)
            except DocumentMismatchError as e:
                _LOG.warning("%s: Skipping", str(e))
                b_skipped += 1
            except Exception as e:
                _LOG.warning("%s: Skipping", str(e))
                b_skipped += 1
        return BatchStatus(b_added, b_skipped, monotonic() - b_started, b_loaded)

    def bulk_add(self,
                 metadata_docs: Iterable[JsonDict],
                 batch_size: int = 1000) -> BatchStatus:
        """
        Add a group of Metadata Type documents in bulk.

        API Note: This API method is not finalised and may be subject to change.

        :param metadata_docs: An iterable of metadata type metadata docs.
        :param batch_size: Number of metadata types to add per batch (default 1000)
        :return: BatchStatus named tuple, with `safe` containing a list of
                 metadata type names that are safe to include in a subsequent product bulk add.
        """
        n_in_batch = 0
        added = 0
        skipped = 0
        started = monotonic()
        batch = []
        existing = {mdt.name: mdt for mdt in self.get_all()}
        batched = set()
        safe = set()
        for doc in metadata_docs:
            try:
                mdt = self.from_doc(doc)
                if mdt.name in existing:
                    check_doc_unchanged(
                        existing[mdt.name].definition,
                        jsonify_document(mdt.definition),
                        'Metadata Type {}'.format(mdt.name)
                    )
                    _LOG.warning("%s: Skipped - already exists", mdt.name)
                    skipped += 1
                    safe.add(mdt.name)
                else:
                    batch.append(mdt)
                    batched.add(mdt.name)
                    n_in_batch += 1
            except DocumentMismatchError as e:
                _LOG.warning("%s: Skipped", str(e))
                skipped += 1
            except InvalidDocException as e:
                _LOG.warning("%s: Skipped", str(e))
                skipped += 1
            if n_in_batch >= batch_size:
                batch_results = self._add_batch(batch)
                batch = []
                added += batch_results.completed
                skipped += batch_results.skipped
                if batch_results.safe is None:
                    safe.update(batched)
                else:
                    safe.update(batch_results.safe)
                batched = set()
                n_in_batch = 0
        if n_in_batch > 0:
            batch_results = self._add_batch(batch)
            added += batch_results.completed
            skipped += batch_results.skipped
            if batch_results.safe is None:
                safe.update(batched)
            else:
                safe.update(batch_results.safe)
        return BatchStatus(added, skipped, monotonic() - started, safe)

    @abstractmethod
    def can_update(self,
                   metadata_type: MetadataType,
                   allow_unsafe_updates: bool = False
                  ) -> tuple[bool, Iterable[Change], Iterable[Change]]:
        """
        Check if metadata type can be updated. Return bool,safe_changes,unsafe_changes

        Safe updates currently allow new search fields to be added, description to be changed.

        :param metadata_type: updated MetadataType
        :param allow_unsafe_updates: Allow unsafe changes. Use with caution.
        :return: Tuple of: boolean (can/can't update); safe changes; unsafe changes
        """

    @abstractmethod
    def update(self,
               metadata_type: MetadataType,
               allow_unsafe_updates: bool = False,
               allow_table_lock: bool = False
              ) -> MetadataType:
        """
        Update a metadata type from the document. Unsafe changes will throw a ValueError by default.

        Safe updates currently allow new search fields to be added, description to be changed.

        :param metadata_type: MetadataType model with unpersisted updates
        :param allow_unsafe_updates: Allow unsafe changes. Use with caution.
        :param allow_table_lock:
            Allow an exclusive lock to be taken on the table while creating the indexes.
            This will halt other user's requests until completed.

            If false, creation will be slower and cannot be done in a transaction.
        :return: Persisted updated MetadataType model
        """

    def update_document(self,
                        definition: JsonDict,
                        allow_unsafe_updates: bool = False,
                        ) -> MetadataType:
        """
        Update a metadata type from the document. Unsafe changes will throw a ValueError by default.

        Safe updates currently allow new search fields to be added, description to be changed.

        :param definition: Updated definition
        :param allow_unsafe_updates: Allow unsafe changes. Use with caution.
        :return: Persisted updated MetadataType model
        """
        return self.update(self.from_doc(definition), allow_unsafe_updates=allow_unsafe_updates)

    def get_with_fields(self, field_names: Iterable[str]) -> Iterable[MetadataType]:
        """
        Return all metadata types that have all of the named search fields.

        :param field_names: Iterable of search field names
        :return: Iterable of matching metadata types.
        """
        for mdt in self.get_all():
            if all(field in mdt.dataset_fields for field in field_names):
                yield mdt

    def get(self, id_: int) -> MetadataType | None:
        """
        Fetch metadata type by id.

        :return: MetadataType model or None if not found
        """
        try:
            return self.get_unsafe(id_)
        except KeyError:
            return None

    def get_by_name(self, name: str) -> MetadataType | None:
        """
        Fetch metadata type by name.

        :return: MetadataType model or None if not found
        """
        try:
            return self.get_by_name_unsafe(name)
        except KeyError:
            return None

    @abstractmethod
    def get_unsafe(self, id_: int) -> MetadataType:
        """
        Fetch metadata type by id

        :param id_:
        :return: metadata type model
        :raises KeyError: if not found
        """

    @abstractmethod
    def get_by_name_unsafe(self, name: str) -> MetadataType:
        """
        Fetch metadata type by name

        :param name:
        :return: metadata type model
        :raises KeyError: if not found
        """

    @abstractmethod
    def check_field_indexes(self,
                            allow_table_lock: bool = False,
                            rebuild_views: bool = False,
                            rebuild_indexes: bool = False
                           ) -> None:
        """
        Create or replace per-field indexes and views.

        May have no effect if not relevant for this index implementation

        :param allow_table_lock:
            Allow an exclusive lock to be taken on the table while creating the indexes.
            This will halt other user's requests until completed.

            If false, creation will be slightly slower and cannot be done in a transaction.
        :param: rebuild_views: whether or not views should be rebuilt
        :param: rebuild_indexes: whether or not views should be rebuilt
        """

    @abstractmethod
    def get_all(self) -> Iterable[MetadataType]:
        """
        Retrieve all Metadata Types

        :returns: All available MetadataType models
        """

    def get_all_docs(self) -> Iterable[JsonDict]:
        """
        Retrieve all Metadata Types as documents only (e.g. for an index clone)

        Default implementation calls self.get_all()

        API Note: This API method is not finalised and may be subject to change.

        :returns: All available MetadataType definition documents
        """
        # Default implementation calls get_all()
        for mdt in self.get_all():
            yield cast(JsonDict, mdt.definition)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Custom types for postgres & sqlalchemy
"""

from sqlalchemy import TIMESTAMP, text
from sqlalchemy.types import Double
from sqlalchemy.dialects.postgresql.ranges import AbstractRange, Range
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql import sqltypes
from sqlalchemy.sql.expression import Executable, ClauseElement
from sqlalchemy.sql.functions import GenericFunction

SCHEMA_NAME = 'agdc'


class CreateView(Executable, ClauseElement):
    inherit_cache = True

    def __init__(self, name, select):
        self.name = name
        self.select = select


@compiles(CreateView)
def visit_create_view(element, compiler, **kw):
    return "CREATE VIEW %s AS %s" % (
        element.name,
        compiler.process(element.select, literal_binds=True)
    )


UPDATE_TIMESTAMP_SQL = """
create or replace function {schema}.set_row_update_time()
returns trigger as $$
begin
  new.updated = now();
  return new;
end;
$$ language plpgsql;
""".format(schema=SCHEMA_NAME)

UPDATE_COLUMN_MIGRATE_SQL_TEMPLATE = """
alter table {schema}.{table} add column if not exists updated
timestamptz default now();
"""

UPDATE_COLUMN_INDEX_SQL_TEMPLATE = """
create index if not exists ix_{table}_updated
on {schema}.{table}(updated);
"""

ADDED_COLUMN_MIGRATE_SQL_TEMPLATE = """
alter table {schema}.{table} add column if not exists added
timestamptz default now();
"""

ADDED_COLUMN_INDEX_SQL_TEMPLATE = """
create index if not exists ix_{table}_added
on {schema}.{table}(added);
"""

INSTALL_TRIGGER_SQL_TEMPLATE = """
drop trigger if exists row_update_time_{table} on {schema}.{table};
create trigger row_update_time_{table}
before update on {schema}.{table}
for each row
execute procedure {schema}.set_row_update_time();
"""

TYPES_INIT_SQL = """
create or replace function {schema}.common_timestamp(text)
returns timestamp with time zone as $$
select ($1)::timestamp at time zone 'utc';
$$ language sql immutable returns null on null input;

create type {schema}.float8range as range (
    subtype = float8,
    subtype_diff = float8mi
);
""".format(schema=SCHEMA_NAME)


# pylint: disable=abstract-method
class FLOAT8RANGE(AbstractRange[Range[Double]]):
    __visit_name__ = 'FLOAT8RANGE'


@compiles(FLOAT8RANGE)
def visit_float8range(element, compiler, **kw):
    return "FLOAT8RANGE"


# Register the function with SQLAlchemhy.
# pylint: disable=too-many-ancestors
class CommonTimestamp(GenericFunction):
    type = TIMESTAMP(timezone=True)
    package = 'agdc'
    identifier = 'common_timestamp'
    inherit_cache = False

    name = 'common_timestamp'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.packagenames = ['%s' % SCHEMA_NAME]


# pylint: disable=too-many-ancestors
class Float8Range(GenericFunction):
    type = FLOAT8RANGE  # type: ignore[assignment]
    package = 'agdc'
    identifier = 'float8range'
    inherit_cache = False

    name = 'float8range'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.packagenames = ['%s' % SCHEMA_NAME]


class PGNAME(sqltypes.Text):
    """Postgres 'NAME' type."""
    __visit_name__ = 'NAME'


@compiles(PGNAME)
def visit_name(element, compiler, **kw):
    return "NAME"


def pg_exists(conn, name):
    """
    Does a postgres object exist?
    :rtype bool
    """
    return conn.execute(text(f"SELECT to_regclass('{name}')")).scalar() is not None


def pg_column_exists(conn, table, column):
    """
    Does a postgres object exist?
    :rtype bool
    """
    return conn.execute(text(f"""
                        SELECT 1 FROM pg_attribute
                        WHERE attrelid = to_regclass('{table}')
                        AND attname = '{column}'
                        AND NOT attisdropped
                        """)).scalar() is not None


def escape_pg_identifier(conn, name):
    """
    Escape identifiers (tables, fields, roles, etc) for inclusion in SQL statements.

    psycopg2 can safely merge query arguments, but cannot do the same for dynamically
    generating queries.

    See http://initd.org/psycopg/docs/sql.html for more information.
    """
    # New (2.7+) versions of psycopg2 have function: extensions.quote_ident()
    # But it's too bleeding edge right now. We'll ask the server to escape instead, as
    # these are not performance sensitive.
    return conn.execute(text(f"select quote_ident('{name}')")).scalar()



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
High level indexing operations/utilities
"""
import logging

import json
import toolz
from uuid import UUID
from typing import cast, Any, Callable, Optional, Iterable, List, Mapping, Sequence, Tuple, Union, MutableMapping

from datacube.model import Dataset, LineageTree, Product
from datacube.index.abstract import AbstractIndex
from datacube.utils import changes, InvalidDocException, SimpleDocNav, jsonify_document
from datacube.model.utils import BadMatch, dedup_lineage, remap_lineage_doc, flatten_datasets
from datacube.utils.changes import get_doc_changes
from datacube.model import LineageDirection
from .eo3 import prep_eo3, is_doc_eo3, is_doc_geo  # type: ignore[attr-defined]

_LOG = logging.getLogger(__name__)


class ProductRule:
    def __init__(self, product: Product, signature: Mapping[str, Any]):
        self.product = product
        self.signature = signature


def load_rules_from_types(index: AbstractIndex,
                          product_names: Optional[Iterable[str]] = None,
                          excluding: Optional[Iterable[str]] = None
                          ) -> Union[Tuple[List[ProductRule], None], Tuple[None, str]]:
    products: List[Product] = []
    if product_names:
        for name in product_names:
            product = index.products.get_by_name(name)
            if not product:
                return None, 'Supplied product name "%s" not present in the database' % name
            products.append(product)
    else:
        products += index.products.get_all()

    if excluding is not None:
        excluding = set(excluding)
        products = [p for p in products if p.name not in excluding]

    if len(products) == 0:
        return None, 'Found no matching products in the database'

    return [ProductRule(p, p.metadata_doc) for p in products], None


ProductMatcher = Callable[[Mapping[str, Any]], Product]


def product_matcher(rules: Sequence[ProductRule]) -> ProductMatcher:
    """Given product matching rules return a function mapping a document to a
    matching product.

    """
    assert len(rules) > 0

    def matches(doc: Mapping[str, Any], rule: ProductRule) -> bool:
        return changes.contains(doc, rule.signature)

    def single_product_matcher(rule):
        def match(doc: Mapping[str, Any]) -> bool:
            if matches(doc, rule):
                return rule.product

            raise BadMatch('Dataset metadata did not match product signature.'
                           '\nDataset definition:\n %s\n'
                           '\nProduct signature:\n %s\n'
                           % (json.dumps(doc, indent=4),
                              json.dumps(rule.signature, indent=4)))

        return match

    if len(rules) == 1:
        return single_product_matcher(rules[0])

    def match(doc: Mapping[str, Any]) -> Product:
        matched = [rule.product for rule in rules if changes.contains(doc, rule.signature)]

        if len(matched) == 1:
            return matched[0]

        doc_id = doc.get('id', '<missing id>')

        if len(matched) == 0:
            raise BadMatch('No matching Product found for dataset %s' % doc_id)
        else:
            raise BadMatch('Auto match failed, dataset %s matches several products:\n  %s' % (
                doc_id,
                ','.join(p.name for p in matched)))

    return match


def check_dataset_consistent(dataset: Dataset) -> Tuple[bool, Optional[str]]:
    """
    :type dataset: datacube.model.Dataset
    :return: (Is consistent, [error message|None])
    :rtype: (bool, str or None)
    """
    product_measurements = set(dataset.product.measurements.keys())

    if len(product_measurements) == 0:
        return True, None

    if dataset.measurements is None:
        return False, "No measurements defined for a dataset"

    # It the type expects measurements, ensure our dataset contains them all.
    if not product_measurements.issubset(dataset.measurements.keys()):
        # Exclude 3D measurements since it's just a mapping to 2D measurements
        not_measured = {
            m
            for m in product_measurements - set(dataset.measurements.keys())
            if "extra_dim" not in dataset.product.measurements.get(m, [])
        }

        if not_measured:
            msg = "The dataset is not specifying all of the measurements in this product.\n"
            msg += "Missing fields are;\n" + str(not_measured)
            return False, msg

    return True, None


def check_consistent(a: Mapping[str, Any], b: Mapping[str, Any]) -> Tuple[bool, Optional[str]]:
    diffs = get_doc_changes(a, b)
    if len(diffs) == 0:
        return True, None

    def render_diff(offset, a, b):
        offset = '.'.join(map(str, offset))
        return '{}: {!r}!={!r}'.format(offset, a, b)

    return False, ", ".join([render_diff(offset, a, b) for offset, a, b in diffs])


DatasetOrError = Union[
    Tuple[Dataset, None],
    Tuple[None, Union[str, Exception]]
]


def check_intended_eo3(ds: SimpleDocNav, product: Product) -> None:
    # warn if it looks like dataset was meant to be eo3 but is not
    if not is_doc_eo3(ds.doc) and ("eo3" in product.metadata_type.name):
        _LOG.warning(f"Dataset {ds.id} has a product with an eo3 metadata type, "
                     "but the dataset definition does not include the $schema field "
                     "and so will not be recognised as an eo3 dataset.")


def resolve_no_lineage(ds: SimpleDocNav,
                       uri: str,
                       matcher: ProductMatcher,
                       source_tree: Optional[LineageTree] = None,
                       home_index: str | None = None) -> DatasetOrError:
    if source_tree or home_index:
        raise ValueError("source_tree passed to non-lineage resolver")
    doc = ds.doc_without_lineage_sources
    try:
        product = matcher(doc)
    except BadMatch as e:
        return None, e
    check_intended_eo3(ds, product)
    return Dataset(product, doc, uri=uri, sources={}), None


def resolve_with_lineage(doc: SimpleDocNav,
                         uri: str, matcher: ProductMatcher,
                         source_tree: Optional[LineageTree] = None,
                         home_index: Optional[str] = None) -> DatasetOrError:
    """
    Dataset driver for the (new) external lineage API

    API parameters
    :param doc: Dataset docnav
    :param uri: location uri

    Extra kwargs passed in by Doc2Dataset:
    :param matcher: Product matcher
    :param source_tree: sourcewards LineageTree to use in place of EO3 sources (optional)
    :param home_index: Home for sources (ignored if source_tree is not none)
    :return:
    """
    uuid_ = doc.id
    if not uuid_:
        return None, "No id defined in dataset doc"
    try:
        product = matcher(doc.doc)
    except BadMatch as e:
        return None, e
    if source_tree is None:
        # Get sources from EO3 document, use home_index as home of source id's
        source_tree = LineageTree.from_data(uuid_, sources=doc.sources, home=home_index)
    else:
        # May be None
        if source_tree.direction == LineageDirection.DERIVED:
            raise ValueError("source_tree cannot be a derived tree.")
        source_tree = source_tree.find_subtree(uuid_)
    check_intended_eo3(doc, product)
    return Dataset(product,
                   doc.doc,
                   source_tree=source_tree,
                   uri=uri), None


def resolve_legacy_lineage(main_ds_doc: SimpleDocNav, uri: str, matcher: ProductMatcher,
                           index: AbstractIndex,
                           fail_on_missing_lineage: bool,
                           verify_lineage: bool,
                           source_tree: Optional[LineageTree] = None,
                           home_index: str | None = None
                           ) -> DatasetOrError:
    if source_tree or home_index:
        raise ValueError("source_tree passed to non-external lineage resolver")

    try:
        main_ds = SimpleDocNav(dedup_lineage(main_ds_doc))
    except InvalidDocException as e:
        return None, e

    main_uuid = main_ds.id
    if not main_uuid:
        return None, "No id defined in dataset doc"

    ds_by_uuid = toolz.valmap(toolz.first, flatten_datasets(main_ds))
    all_uuid = list(ds_by_uuid)
    db_dss = {ds.id: ds for ds in index.datasets.bulk_get(all_uuid)}

    lineage_uuids = set(filter(lambda x: x != main_uuid, all_uuid))
    missing_lineage = lineage_uuids - set(db_dss)

    if missing_lineage and fail_on_missing_lineage:
        return None, "Following lineage datasets are missing from DB: %s" % (
            ','.join(str(m) for m in missing_lineage))

    if not is_doc_eo3(main_ds.doc):
        if is_doc_geo(main_ds.doc, check_eo3=False):
            if not index.supports_legacy:
                return None, "Legacy metadata formats not supported by the current index driver."
        else:
            if not index.supports_nongeo:
                return None, "Non-geospatial metadata formats not supported by the current index driver."
        if verify_lineage:
            bad_lineage = []

            for uuid in lineage_uuids:
                if uuid in db_dss:
                    ok, err = check_consistent(jsonify_document(ds_by_uuid[uuid].doc_without_lineage_sources),
                                               db_dss[uuid].metadata_doc)
                    if not ok:
                        bad_lineage.append((uuid, err))

            if len(bad_lineage) > 0:
                error_report = '\n'.join('Inconsistent lineage dataset {}:\n> {}'.format(uuid, err)
                                         for uuid, err in bad_lineage)
                return None, error_report

    def with_cache(v: Dataset, k: UUID, cache: MutableMapping[UUID, Dataset]) -> Dataset:
        cache[k] = v
        return v

    def resolve_ds(ds: SimpleDocNav,
                   sources: Optional[Mapping[UUID, Dataset]],
                   cache: MutableMapping[UUID, Dataset]) -> Dataset:
        cached = cache.get(ds.id)
        if cached is not None:
            return cached

        this_uri = uri if ds.id == main_uuid else None

        doc = ds.doc

        db_ds = db_dss.get(ds.id)
        if db_ds:
            product = db_ds.product
        else:
            product = matcher(doc)

        check_intended_eo3(ds, product)
        return with_cache(Dataset(product, doc, uri=this_uri, sources=sources), ds.id, cache)
    try:
        return remap_lineage_doc(main_ds, resolve_ds, cache={}), None
    except BadMatch as e:
        return None, e


def dataset_resolver(index: AbstractIndex,
                     match_product: Callable[[Mapping[str, Any]], Product],
                     fail_on_missing_lineage: bool = False,
                     verify_lineage: bool = True,
                     skip_lineage: bool = False,
                     home_index: Optional[str] = None) -> Callable[[SimpleDocNav, str, LineageTree | None],
                                                                   DatasetOrError
                                                                   ]:
    if skip_lineage or not index.supports_lineage:
        # Resolver that ignores lineage.
        resolver: Callable[..., DatasetOrError] = resolve_no_lineage
        extra_kwargs: dict[str, Any] = {
            "matcher": match_product,
        }
    elif index.supports_external_lineage:
        # ODCv2 external lineage API resolver
        resolver = resolve_with_lineage
        extra_kwargs = {
            "matcher": match_product,
            "home_index": home_index,
        }
    else:
        # Legacy lineage API resolver
        resolver = resolve_legacy_lineage
        extra_kwargs = {
            "matcher": match_product,
            "index": index,
            "fail_on_missing_lineage": fail_on_missing_lineage,
            "verify_lineage": verify_lineage,
        }

    def resolve(doc: SimpleDocNav, uri: str, source_tree: Optional[LineageTree] = None) -> DatasetOrError:
        return resolver(doc, uri, source_tree=source_tree, **extra_kwargs)

    return resolve


class Doc2Dataset:
    """Used for constructing `Dataset` objects from plain metadata documents.

    This requires a database connection to perform the automatic matching against
    available products.

    There are options for including and excluding the products to match against,
    as well as how to deal with source lineage.

    Once constructed, call with a dictionary object and location URI, eg::

        resolver = Doc2Dataset(index)
        dataset = resolver(dataset_dictionary, 'file:///tmp/test-dataset.json')
        index.dataset.add(dataset)


    :param index: an open Database connection

    :param list products: List of product names against which to match datasets
                          (including lineage datasets). If not supplied we will
                          consider all products.

    :param list exclude_products: List of products to exclude from matching

    :param fail_on_missing_lineage: If True fail resolve if any lineage
                                    datasets are missing from the DB

                                    Only False supported if index.supports_external_lineage is True.

    :param verify_lineage: If True check that lineage datasets in the
                           supplied document are identical to DB versions

                           Ignored for EO3 documents.  Will be dropped in ODCv2 as only eo3 documents
                           will be supported.

    :param skip_lineage: If True ignore lineage sub-tree in the supplied
                         document and construct dataset without lineage datasets
    :param eo3: 'auto'/True/False by default auto-detect EO3 datasets and pre-process them

                Cannot be 'False' if index.supports_legacy is False.
                Will be dropped in ODCv2 as only eo3 documents will be supported
    :param home_index: Ignored if index.supports_exernal_home is False.  Defaults to None.
                Optional string labelling the "home index" for lineage datasets.
                home_index is ignored if an explicit source_tree is passed to the resolver.
    """
    def __init__(self,
                 index: AbstractIndex,
                 products: Optional[Sequence[str]] = None,
                 exclude_products: Optional[Sequence[str]] = None,
                 fail_on_missing_lineage: bool = False,
                 verify_lineage: bool = True,
                 skip_lineage: bool = False,
                 eo3: Union[bool, str] = 'auto',
                 home_index: Optional[str] = None):
        if not index.supports_lineage:
            skip_lineage = True
            verify_lineage = False
            fail_on_missing_lineage = False
            home_index = None
        else:
            if not index.supports_legacy and not index.supports_nongeo:
                if not eo3:
                    raise ValueError("EO3 cannot be set to False for a non-legacy geo-only index.")
                eo3 = True
            if index.supports_external_lineage and fail_on_missing_lineage:
                raise ValueError("fail_on_missing_lineage is not supported for this index driver.")
            if home_index and skip_lineage:
                raise ValueError("Cannot provide a default home_index when skip_lineage is set.")

        rules, err_msg = load_rules_from_types(index,
                                               product_names=products,
                                               excluding=exclude_products)
        if rules is None:
            raise ValueError(err_msg)

        self.index = index
        self._eo3 = eo3
        matcher = product_matcher(rules)
        self._ds_resolve = dataset_resolver(index,
                                            matcher,
                                            fail_on_missing_lineage=fail_on_missing_lineage,
                                            verify_lineage=verify_lineage,
                                            skip_lineage=skip_lineage,
                                            home_index=home_index)

    def __call__(self, doc_in: Union[SimpleDocNav, Mapping[str, Any]], uri: str,
                 source_tree: Optional[LineageTree] = None) -> DatasetOrError:
        """Attempt to construct dataset from metadata document and a uri.

        :param doc: Dictionary or SimpleDocNav object
        :param uri: String "location" property of the Dataset

        :return: (dataset, None) is successful,
        :return: (None, ErrorMessage) on failure
        """
        if isinstance(doc_in, SimpleDocNav):
            doc: SimpleDocNav = doc_in
        else:
            doc = SimpleDocNav(doc_in)

        if self._eo3:
            auto_skip = self._eo3 == 'auto'
            doc = SimpleDocNav(
                prep_eo3(
                    doc.doc,
                    auto_skip=auto_skip,
                    remap_lineage=not self.index.supports_external_lineage
                )
            )

        dataset, err = self._ds_resolve(doc, uri, source_tree)
        if dataset is None:
            return None, cast(Union[str, Exception], err)

        is_consistent, reason = check_dataset_consistent(dataset)
        if not is_consistent:
            return None, cast(Union[str, Exception], reason)

        return dataset, None



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from uuid import uuid4 as random_uuid

import pytest

from datacube.model import LineageDirection, LineageTree, InconsistentLineageException
from datacube.model.lineage import LineageRelations


@pytest.mark.parametrize('datacube_env_name', ('experimental',))
def test_lineage_home_api(index):
    a_uuids = [random_uuid() for i in range(10)]
    b_uuids = [random_uuid() for i in range(10)]
    all_uuids = a_uuids + b_uuids
    assert index.lineage.get_homes(*a_uuids) == {}
    # Test delete of non-existent entries
    assert index.lineage.clear_home(*a_uuids) == 0
    # Test insert a uuids
    assert index.lineage.set_home("spam", *a_uuids) == 10
    for home in index.lineage.get_homes(*a_uuids).values():
        assert home == "spam"
    # Test update with and without allow_update
    index.lineage.set_home("eggs", *a_uuids) == 0
    index.lineage.set_home("eggs", *a_uuids, allow_updates=True) == 10
    for home in index.lineage.get_homes(*a_uuids).values():
        assert home == "eggs"
    assert index.lineage.get_homes(*b_uuids) == {}
    index.lineage.set_home("eggs", *a_uuids, allow_updates=True) == 0
    index.lineage.set_home("eggs", *b_uuids, allow_updates=True) == 10

    # Test clear_home with actual work done.
    assert index.lineage.clear_home(*a_uuids) == 10
    assert index.lineage.clear_home(*b_uuids) == 10


@pytest.mark.parametrize('datacube_env_name', ('experimental',))
def test_lineage_merge(index, src_lineage_tree, compatible_derived_tree):
    stree, ids = src_lineage_tree
    dtree, ids = compatible_derived_tree

    rels = LineageRelations(tree=stree)
    rels.merge_tree(dtree)
    index.lineage.merge(rels)
    src_tree = index.lineage.get_source_tree(ids["root"])
    assert src_tree.dataset_id == ids["root"]
    assert src_tree.direction == LineageDirection.SOURCES
    for ard_subtree in src_tree.children["ard"]:
        assert ard_subtree.dataset_id in (ids["ard1"], ids["ard2"])


@pytest.mark.parametrize('datacube_env_name', ('experimental',))
def test_lineage_tree_index_api_simple(index, src_lineage_tree):
    tree, ids = src_lineage_tree
    # Test api responses for lineage not in database:
    src_tree = index.lineage.get_source_tree(ids["root"])
    assert src_tree.dataset_id == ids["root"]
    assert src_tree.direction == LineageDirection.SOURCES
    assert src_tree.children == {}
    # Add the test tree to depth 1
    index.lineage.add(tree, max_depth=1)
    src_tree = index.lineage.get_source_tree(ids["root"])
    assert src_tree.dataset_id == ids["root"]
    assert src_tree.direction == LineageDirection.SOURCES
    for ard_subtree in src_tree.children["ard"]:
        assert ard_subtree.dataset_id in (ids["ard1"], ids["ard2"])
        assert not ard_subtree.children
    # Add the test tree to depth 2
    index.lineage.add(tree, max_depth=2)
    src_tree = index.lineage.get_source_tree(ids["root"])
    for ard_subtree in src_tree.children["ard"]:
        assert "l1" in ard_subtree.children
        assert not ard_subtree.children["atmos_corr"][0].children
    # Add full test tree
    index.lineage.add(tree, max_depth=0)
    src_tree = index.lineage.get_source_tree(ids["root"])
    seen = False
    for ard_subtree in src_tree.children["ard"]:
        assert "l1" in ard_subtree.children
        assert "atmos_corr" in ard_subtree.children
        if ard_subtree.children["atmos_corr"][0].children:
            assert "preatmos" in ard_subtree.children["atmos_corr"][0].children
            seen = True
    assert seen
    # And test reversing the tree
    der_tree = index.lineage.get_derived_tree(ids["atmos_parent"])
    assert der_tree.find_subtree(ids["root"]).dataset_id == ids["root"]
    # Test Lineage removal - sourcewards
    index.lineage.remove(ids["root"], LineageDirection.SOURCES, max_depth=2)
    src_tree = index.lineage.get_source_tree(ids["root"])
    assert not src_tree.children
    src_tree = index.lineage.get_source_tree(ids["atmos"])
    assert src_tree.children
    # Test Lineage removal - derivedwards
    index.lineage.add(tree, max_depth=0)
    index.lineage.remove(ids["atmos_parent"], LineageDirection.DERIVED, max_depth=2)
    src_tree = index.lineage.get_source_tree(ids["root"])
    assert src_tree.children
    src_tree = index.lineage.get_source_tree(ids["atmos"])
    assert not src_tree.children


@pytest.mark.parametrize('datacube_env_name', ('experimental',))
def test_lineage_tree_index_api_consistent(index, src_lineage_tree, compatible_derived_tree):
    tree1, ids = src_lineage_tree
    tree2, ids = compatible_derived_tree

    index.lineage.add(tree1)
    tree1a = index.lineage.get_source_tree(ids["root"])
    assert tree1a.home is None

    index.lineage.add(tree2)
    tree2a = index.lineage.get_source_tree(ids["root"])
    assert tree2a.home == "extensions"
    tree2b = tree2a.find_subtree(ids["atmos"])
    tree2c = tree2b.find_subtree(ids["atmos_grandparent"])
    assert tree2c


@pytest.mark.parametrize('datacube_env_name', ('experimental',))
def test_lineage_tree_index_api_inconsistent_homes(index, src_lineage_tree):
    tree, ids = src_lineage_tree
    home_update = LineageTree(
        dataset_id=ids["ard1"],
        direction=LineageDirection.SOURCES,
        home="not_too_ard",
        children={
            "l1": [
                LineageTree(
                    dataset_id=ids["l1_1"],
                    direction=LineageDirection.SOURCES,
                    home="level1a",
                )
            ]
        }
    )
    index.lineage.add(tree)
    with pytest.raises(InconsistentLineageException):
        index.lineage.add(home_update, allow_updates=False)
    index.lineage.add(home_update, allow_updates=True)
    dbtree = index.lineage.get_source_tree(ids["ard1"])
    assert dbtree.home == "not_too_ard"


@pytest.mark.parametrize('datacube_env_name', ('experimental',))
def test_get_extensions(index, dataset_with_external_lineage):
    dataset, src_lineage_tree, derived_lineage_tree, ids = dataset_with_external_lineage

    ds = index.datasets.get(ids["root"])
    assert ds.source_tree is None
    assert ds.derived_tree is None

    ds = index.datasets.get(ids["root"], include_sources=True)
    assert ds.source_tree is not None
    assert ds.derived_tree is None
    assert ds.source_tree.children["ard"][0].children

    ds = index.datasets.get(ids["root"], include_sources=True, max_depth=1)
    assert ds.source_tree is not None
    assert ds.derived_tree is None
    assert not ds.source_tree.children["ard"][0].children

    ds = index.datasets.get(ids["root"], include_deriveds=True)
    assert ds.source_tree is None
    assert ds.derived_tree is not None
    assert ds.derived_tree.children["dra"][0].children

    ds = index.datasets.get(ids["root"], include_deriveds=True, max_depth=1)
    assert ds.source_tree is None
    assert ds.derived_tree is not None
    assert not ds.derived_tree.children["dra"][0].children

    ds = index.datasets.get(ids["root"], include_sources=True, include_deriveds=True)
    assert ds.source_tree is not None
    assert ds.derived_tree is not None
    assert ds.source_tree.children["ard"][0].children
    assert ds.derived_tree.children["dra"][0].children

    ds = index.datasets.get(ids["root"], include_sources=True, include_deriveds=True, max_depth=1)
    assert ds.source_tree is not None
    assert ds.derived_tree is not None
    assert not ds.source_tree.children["ard"][0].children
    assert not ds.derived_tree.children["dra"][0].children



import sys

from time import monotonic
from datacube import Datacube
from datacube.model import Range
from datetime import timezone
from datetime import datetime
from odc.geo.geom import CRS, polygon


def benchmark(test, dc, label, n):
    total = 0.0
    total_first = 0.0
    last_count = None
    for i in range(n):
        start = monotonic()
        # count, first = test(dc)
        count = 0
        first = None
        for ds in test(dc):
            if not count:
                first = monotonic()
            count += 1
        if count == 0:
            first = start
        end = monotonic()
        if last_count and count != last_count:
            print(f"Count mismatch in {label}: {count} vs {last_count}")
        last_count = count
        print(f"Test {label}#{i+1}: {end-start}s   ({first-start}s to first returned dataset)")
        total += end - start
        total_first += first - start
    print(f"Test {label}-count: {count} rows")
    print(f"Test {label}-avg: {total/n}s  ({total/(n*count)})s/row")
    print(f"Test {label}-avg-to-first-return: {total_first/n}s  ({total/(n*count)})s/row")
    print()
    print("-----------------------------------------------------------------")


def test_less_than(dc):
    return dc.index.datasets.search(
        product='ga_ls8c_ard_3',
        cloud_cover=Range(None, 0.2)
    )


def test_geospatial_search(dc):
    return dc.index.datasets.search(
        product='ga_ls8c_ard_3',
        lat=Range(-30.0, -25.0),
        lon=Range(140.0, 145.0),
    )


def test_offset_geom(dc):
    if dc.index.supports_external_lineage:
        return dc.index.datasets.search(
            product='ga_ls8c_ard_3',
            geometry=polygon(
                [
                    [140.0, -25.0],
                    [142.0, -25.0],
                    [145.0, -30.0],
                    [145.0, -30.0],
                    [140.0, -25.0],
                ],
                crs=CRS("epsg:4326")
            )
        )
    else:
        return dc.index.datasets.search(
            product='ga_ls8c_ard_3',
            lat=Range(-30.0, -25.0),
            lon=Range(140.0, 145.0),
        )


def test_temporal_search(dc):
    return dc.index.datasets.search(
        product='ga_ls8c_ard_3',
        time=Range(datetime(2016, 1, 1, tzinfo=timezone.utc), datetime(2016, 4, 5, tzinfo=timezone.utc)),
    )


def main(args):
    if args:
        env = args.pop()
    else:
        env = "datacube_real"
    print("Testing on database ", env)
    dc = Datacube(env=env)
    benchmark(test_less_than, dc, "less_than", 20)
    benchmark(test_geospatial_search, dc, "geospatial", 20)
    benchmark(test_temporal_search, dc, "temporal", 20)
    benchmark(test_offset_geom, dc, "geom", 20)


if __name__ == "__main__":
    args = sys.argv[1:]
    main(args)

# For custom CRS search tests (TODO)
# BoundingBox(
#       left=762759.2567816022, bottom=-3326371.8490792206,
#       right=1295116.9248742603, top=-2727561.09954842, crs=CRS('EPSG:3577'))



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Low level config path resolution, loading and multi-format parsing functions.

The default search path and default config also live here.
"""

import os
import warnings
from enum import Enum
from os import PathLike
from os.path import expanduser

from datacube.cfg.exceptions import ConfigException
from datacube.cfg.utils import ConfigDict, smells_like_ini, SemaphoreCallback

_DEFAULT_CONFIG_SEARCH_PATH = [
    "datacube.conf",      # i.e. in the current working directory.
    expanduser("~/.datacube.conf"),   # i.e. in user's home directory.
    # Check if we are running under a Windows and use Windowsy default paths?
    "/etc/default/datacube.conf",  # Preferred location for global config file
    "/etc/datacube.conf",          # Legacy location for global config file
]

_DEFAULT_CONF = """
default:
   db_hostname: ''
   db_database: datacube
   index_driver: default
   db_connection_timeout: 60
"""


def find_config(paths_in: None | str | PathLike | list[str | PathLike],
                default_cb: SemaphoreCallback | None = None) -> str:
    """
    Given a file system path, or a list of file system paths, return the contents of the first file
    in the list that can be read as a string.

    If "None" is passed in the default config search path is used:

        "datacube.conf"               (i.e. 'datacube.conf' in the current working directory.)
        "~/.datacube.conf"            (i.e. '.datacube.conf' in the user's home directory.)
        "/etc/default/datacube.conf"  (Preferred location for global config file)
        "/etc/datacube.conf"          (Legacy location for global config file)

    If a path or list of paths was passed in, AND no readable file could be found, a ConfigException is raised.
    If None was passed in, AND no readable file could be found, a default configuration text is returned.

    :param paths_in: A file system path, or a list of file system paths, or None.
    :return: The contents of the first readable file found.
    """
    using_default_paths: bool = False
    paths: list[str | PathLike] = []
    if paths_in is None:
        if os.environ.get("ODC_CONFIG_PATH"):
            paths.extend(os.environ["ODC_CONFIG_PATH"].split(':'))
        elif os.environ.get("DATACUBE_CONFIG_PATH"):
            warnings.warn(
                "Datacube config path being determined by legacy $DATACUBE_CONFIG_PATH environment variable. "
                "This environment variable is deprecated and the behaviour of it has changed somewhat since datacube "
                "1.8.x.   Please refer to the documentation for details and switch to $ODC_CONFIG_PATH"
            )
            paths.extend(os.environ["DATACUBE_CONFIG_PATH"].split(':'))
        else:
            paths.extend(_DEFAULT_CONFIG_SEARCH_PATH)
            using_default_paths = True
    elif isinstance(paths_in, str) or isinstance(paths_in, PathLike):
        paths.append(paths_in)
    else:
        paths.extend(paths_in)

    for path in paths:
        try:
            with open(path, "r") as fp:
                return fp.read()
        except PermissionError:
            continue
        except FileNotFoundError:
            continue

    if using_default_paths:
        if default_cb is not None:
            default_cb()
        return _DEFAULT_CONF

    raise ConfigException("No configuration file found in the provided locations")


class CfgFormat(Enum):
    """
    An Enum class for config file formats.
    """
    AUTO = 0   # Format unspecified - autodetect
    INI = 1
    YAML = 2
    JSON = 2   # JSON is a subset of YAML


def parse_text(cfg_text: str, fmt: CfgFormat = CfgFormat.AUTO) -> ConfigDict:
    """
    Parse a string of text in INI, JSON or YAML format into a raw dictionary.

    Raises a ConfigException if the file cannot be parsed.

    :param cfg_text: Configuration string in INI, YAML or JSON format
    :param fmt: Whether to use the ini or yaml/json parser. By default autodetects file format.
    :return: A raw config dictionary
    """
    raw_config = {}
    if fmt == fmt.INI or (
            fmt == fmt.AUTO and smells_like_ini(cfg_text)):
        # INI parsing
        import configparser
        try:
            ini_config = configparser.ConfigParser()
            ini_config.read_string(cfg_text)
            for section in ini_config.sections():
                sect = {}
                for key, value in ini_config.items(section):
                    sect[key] = value

                raw_config[section] = sect
        except configparser.Error as e:
            raise ConfigException(f"Invalid INI file: {e}")
    else:
        # YAML/JSON parsing
        import yaml
        try:
            raw_config = yaml.load(cfg_text, Loader=yaml.Loader)
        except yaml.parser.ParserError as e:
            raise ConfigException(f"Invalid YAML file:{e}")

    return raw_config



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Common methods for index integration tests.
"""
import itertools
import os
from copy import copy, deepcopy
from datetime import timedelta
from pathlib import Path
from types import SimpleNamespace
from typing import Union
from uuid import uuid4

import pytest
import yaml
from click.testing import CliRunner
from hypothesis import HealthCheck, settings
from sqlalchemy import text

import datacube.scripts.cli_app
import datacube.utils
from datacube.drivers.postgres import _core as pgres_core
from datacube.drivers.postgis import _core as pgis_core
from datacube.index import index_connect
from datacube.index.abstract import default_metadata_type_docs
from integration_tests.utils import _make_geotiffs, _make_ls5_scene_datasets, load_yaml_file, \
    GEOTIFF, load_test_products

from datacube.cfg import ODCConfig, ODCEnvironment, psql_url_from_config
from datacube.drivers.postgres import PostgresDb
from datacube.drivers.postgis import PostGisDb
from datacube.model import LineageTree, LineageDirection

_SINGLE_RUN_CONFIG_TEMPLATE = """

"""

INTEGRATION_TESTS_DIR = Path(__file__).parent

_EXAMPLE_LS5_NBAR_DATASET_FILE = INTEGRATION_TESTS_DIR / 'example-ls5-nbar.yaml'

#: Number of time slices to create in sample data
NUM_TIME_SLICES = 3

PROJECT_ROOT = Path(__file__).parents[1]
CONFIG_SAMPLES = PROJECT_ROOT / 'docs' / 'config_samples'

CONFIG_FILE_PATHS = [
    os.path.expanduser('~/.datacube_integration.conf'),
    str(INTEGRATION_TESTS_DIR / 'integration.conf'),
]

# Configure Hypothesis to allow slower tests, because we're testing datasets
# and disk IO rather than scalar values in memory.  Ask @Zac-HD for details.
settings.register_profile(
    'opendatacube', deadline=5000, max_examples=10,
    suppress_health_check=[HealthCheck.too_slow]
)
settings.load_profile('opendatacube')

EO3_TESTDIR = INTEGRATION_TESTS_DIR / 'data' / 'eo3'


def get_eo3_test_data_doc(path):
    from datacube.utils import read_documents
    for path, doc in read_documents(EO3_TESTDIR / path):
        return doc


@pytest.fixture
def ext_eo3_mdt_path():
    return str(EO3_TESTDIR / "eo3_landsat_ard.odc-type.yaml")


@pytest.fixture
def eo3_product_paths():
    return [
        str(EO3_TESTDIR / "ard_ls8.odc-product.yaml"),
        str(EO3_TESTDIR / "ga_ls_wo_3.odc-product.yaml"),
        str(EO3_TESTDIR / "s2_africa_product.yaml"),
    ]


@pytest.fixture
def eo3_dataset_paths():
    return [
        str(EO3_TESTDIR / "ls8_dataset.yaml"),
        str(EO3_TESTDIR / "ls8_dataset2.yaml"),
        str(EO3_TESTDIR / "ls8_dataset3.yaml"),
        str(EO3_TESTDIR / "ls8_dataset4.yaml"),
        str(EO3_TESTDIR / "wo_dataset.yaml"),
        str(EO3_TESTDIR / "s2_africa_dataset.yaml"),
        str(EO3_TESTDIR / "s2_africa_dataset2.yaml"),
    ]


@pytest.fixture
def eo3_dataset_update_path():
    return str(EO3_TESTDIR / "ls8_dataset_update.yaml")


@pytest.fixture
def dataset_with_lineage_doc():
    return (
        get_eo3_test_data_doc("wo_ds_with_lineage.odc-metadata.yaml"),
        's3://dea-public-data/derivative/ga_ls_wo_3/1-6-0/090/086/2016/05/12/'
        'ga_ls_wo_3_090086_2016-05-12_final.stac-item.json'
    )


@pytest.fixture
def eo3_ls8_dataset_doc():
    return (
        get_eo3_test_data_doc("ls8_dataset.yaml"),
        's3://dea-public-data/baseline/ga_ls8c_ard_3/090/086/2016/05/12/'
        'ga_ls8c_ard_3-0-0_090086_2016-05-12_final.stac-item.json'
    )


@pytest.fixture
def eo3_ls8_dataset2_doc():
    return (
        get_eo3_test_data_doc("ls8_dataset2.yaml"),
        's3://dea-public-data/baseline/ga_ls8c_ard_3/090/086/2016/05/28/'
        'ga_ls8c_ard_3-0-0_090086_2016-05-28_final.stac-item.json'
    )


@pytest.fixture
def eo3_ls8_dataset3_doc():
    return (
        get_eo3_test_data_doc("ls8_dataset3.yaml"),
        's3://dea-public-data/baseline/ga_ls8c_ard_3/101/077/2013/04/04/'
        'ga_ls8c_ard_3-0-0_101077_2013-04-04_final.stac-item.json'
    )


@pytest.fixture
def eo3_ls8_dataset4_doc():
    return (
        get_eo3_test_data_doc("ls8_dataset4.yaml"),
        's3://dea-public-data/baseline/ga_ls8c_ard_3/101/077/2013/07/21/'
        'ga_ls8c_ard_3-0-0_101077_2013-07-21_final.stac-item.json'
    )


@pytest.fixture
def eo3_wo_dataset_doc():
    return (
        get_eo3_test_data_doc("wo_dataset.yaml"),
        's3://dea-public-data/derivative/ga_ls_wo_3/1-6-0/090/086/2016/05/12/'
        'ga_ls_wo_3_090086_2016-05-12_final.stac-item.json'
    )


@pytest.fixture
def eo3_africa_dataset_doc():
    return (
        get_eo3_test_data_doc("s2_africa_dataset.yaml"),
        's3://deafrica-sentinel-2/sentinel-s2-l2a-cogs/37/M/CQ/'
        '2022/8/S2A_37MCQ_20220808_0_L2A/S2A_37MCQ_20220808_0_L2A.json'
    )


@pytest.fixture
def eo3_africa_dataset2_doc():
    return (
        get_eo3_test_data_doc("s2_africa_dataset2.yaml"),
        's3://deafrica-sentinel-2/sentinel-s2-l2a-cogs/39/M/XR/'
        '2020/9/S2A_39MXR_20200909_0_L2A/S2A_39MXR_20200909_0_L2A.json'
    )


@pytest.fixture
def datasets_with_unembedded_lineage_doc():
    return [
        (
            get_eo3_test_data_doc("ls8_dataset.yaml"),
            's3://dea-public-data/baseline/ga_ls8c_ard_3/090/086/2016/05/12/'
            'ga_ls8c_ard_3-0-0_090086_2016-05-12_final.stac-item.json'
        ),
        (
            get_eo3_test_data_doc("wo_dataset.yaml"),
            's3://dea-public-data/derivative/ga_ls_wo_3/1-6-0/090/086/2016/05/12/'
            'ga_ls_wo_3_090086_2016-05-12_final.stac-item.json'
        ),
    ]


@pytest.fixture
def extended_eo3_metadata_type_doc():
    return get_eo3_test_data_doc("eo3_landsat_ard.odc-type.yaml")


@pytest.fixture
def eo3_sentinel_metadata_type_doc():
    return get_eo3_test_data_doc("eo3_sentinel_ard.odc-type.yaml")


@pytest.fixture
def extended_eo3_product_doc():
    return get_eo3_test_data_doc("ard_ls8.odc-product.yaml")


@pytest.fixture
def base_eo3_product_doc():
    return get_eo3_test_data_doc("ga_ls_wo_3.odc-product.yaml")


@pytest.fixture
def africa_s2_product_doc():
    return get_eo3_test_data_doc("s2_africa_product.yaml")


@pytest.fixture
def s2_ard_product_doc():
    return get_eo3_test_data_doc("ga_s2am_ard_3.odc-product.yaml")


@pytest.fixture
def final_dataset_doc():
    return (
        get_eo3_test_data_doc("final_dataset.yaml"),
        's3://dea-public-data/baseline/ga_ls8c_ard_3/090/086/2023/04/30'
        'ga_ls8c_ard_3-2-1_090086_2023-04-30_final.stac-item.json'
    )


@pytest.fixture
def nrt_dataset_doc():
    return (
        get_eo3_test_data_doc("nrt_dataset.yaml"),
        's3://dea-public-data/baseline/ga_ls8c_ard_3/090/086/2023/04/30_nrt'
        'ga_ls8c_ard_3-2-1_090086_2023-04-30_nrt.stac-item.json'
    )


@pytest.fixture
def ga_s2am_ard_3_interim_doc():
    return (
        get_eo3_test_data_doc("ga_s2am_ard_3_interim.yaml"),
        's3://dea-public-data/baseline/ga_s2am_ard_3/53/JNN/2021/07/24_interim'
        '20230222T235626/ga_s2am_ard_3-2-1_53JNN_2021-07-24_interim.odc-metadata.yaml'
    )


@pytest.fixture
def ga_s2am_ard_3_final_doc():
    return (
        get_eo3_test_data_doc("ga_s2am_ard_3_final.yaml"),
        's3://dea-public-data/baseline/ga_s2am_ard_3/53/JNN/2021/07/24'
        '20210724T023436/ga_s2am_ard_3-2-1_53JNN_2021-07-24_final.odc-metadata.yaml'
    )


def doc_to_ds(index, product_name, ds_doc, ds_path, src_tree=None, derived_tree=None):
    from datacube.index.hl import Doc2Dataset
    resolver = Doc2Dataset(index, products=[product_name], verify_lineage=False)
    ds, err = resolver(ds_doc, ds_path)
    assert err is None and ds is not None
    if src_tree is not None:
        ds.source_tree = src_tree
    if derived_tree is not None:
        ds.derived_tree = derived_tree
    index.datasets.add(ds, with_lineage=index.supports_lineage)
    return index.datasets.get(ds.id)


def doc_to_ds_no_add(index, product_name, ds_doc, ds_path):
    from datacube.index.hl import Doc2Dataset
    resolver = Doc2Dataset(index, products=[product_name], verify_lineage=False)
    ds, err = resolver(ds_doc, ds_path)
    assert err is None and ds is not None
    return ds


@pytest.fixture
def extended_eo3_metadata_type(index, extended_eo3_metadata_type_doc):
    return index.metadata_types.add(
        index.metadata_types.from_doc(extended_eo3_metadata_type_doc)
    )


@pytest.fixture
def eo3_sentinel_metadata_type(index, eo3_sentinel_metadata_type_doc):
    return index.metadata_types.add(
        index.metadata_types.from_doc(eo3_sentinel_metadata_type_doc)
    )


@pytest.fixture
def ls8_eo3_product(index, extended_eo3_metadata_type, extended_eo3_product_doc):
    return index.products.add_document(extended_eo3_product_doc)


@pytest.fixture
def wo_eo3_product(index, base_eo3_product_doc):
    return index.products.add_document(base_eo3_product_doc)


@pytest.fixture
def africa_s2_eo3_product(index, africa_s2_product_doc):
    return index.products.add_document(africa_s2_product_doc)


@pytest.fixture
def ga_s2am_ard_3_product(index, eo3_sentinel_metadata_type, s2_ard_product_doc):
    return index.products.add_document(s2_ard_product_doc)


@pytest.fixture
def eo3_products(index, extended_eo3_metadata_type,
                 ls8_eo3_product, wo_eo3_product,
                 africa_s2_eo3_product):
    return [
        africa_s2_eo3_product,
        ls8_eo3_product,
        wo_eo3_product,
    ]


@pytest.fixture
def ls8_eo3_dataset(index, extended_eo3_metadata_type, ls8_eo3_product, eo3_ls8_dataset_doc):
    return doc_to_ds(index,
                     ls8_eo3_product.name,
                     *eo3_ls8_dataset_doc)


@pytest.fixture
def ls8_eo3_dataset2(index, extended_eo3_metadata_type, ls8_eo3_product, eo3_ls8_dataset2_doc):
    return doc_to_ds(index,
                     ls8_eo3_product.name,
                     *eo3_ls8_dataset2_doc)


@pytest.fixture
def ls8_eo3_dataset3(index, extended_eo3_metadata_type, ls8_eo3_product, eo3_ls8_dataset3_doc):
    return doc_to_ds(index,
                     ls8_eo3_product.name,
                     *eo3_ls8_dataset3_doc)


@pytest.fixture
def ls8_eo3_dataset4(index, extended_eo3_metadata_type, ls8_eo3_product, eo3_ls8_dataset4_doc):
    return doc_to_ds(index,
                     ls8_eo3_product.name,
                     *eo3_ls8_dataset4_doc)


@pytest.fixture
def wo_eo3_dataset(index, wo_eo3_product, eo3_wo_dataset_doc, ls8_eo3_dataset):
    return doc_to_ds(index,
                     wo_eo3_product.name,
                     *eo3_wo_dataset_doc)


@pytest.fixture
def africa_eo3_dataset(index, africa_s2_eo3_product, eo3_africa_dataset_doc):
    return doc_to_ds(index,
                     africa_s2_eo3_product.name,
                     *eo3_africa_dataset_doc)


@pytest.fixture
def africa_eo3_dataset2(index, africa_s2_eo3_product, eo3_africa_dataset2_doc):
    return doc_to_ds(index,
                     africa_s2_eo3_product.name,
                     *eo3_africa_dataset2_doc)


@pytest.fixture
def nrt_dataset(index, extended_eo3_metadata_type, ls8_eo3_product, nrt_dataset_doc):
    return doc_to_ds_no_add(
        index,
        ls8_eo3_product.name,
        *nrt_dataset_doc)


@pytest.fixture
def final_dataset(index, extended_eo3_metadata_type, ls8_eo3_product, final_dataset_doc):
    return doc_to_ds_no_add(
        index,
        ls8_eo3_product.name,
        *final_dataset_doc)


@pytest.fixture
def ds_no_region(index, extended_eo3_metadata_type, ls8_eo3_product, final_dataset_doc):
    doc_no_region = deepcopy(final_dataset_doc)
    doc_no_region[0]["properties"]["odc:region_code"] = None
    return doc_to_ds_no_add(
        index,
        ls8_eo3_product.name,
        *doc_no_region)


@pytest.fixture
def ds_with_lineage(index, wo_eo3_product, eo3_wo_dataset_doc, ls8_eo3_dataset):
    doc, path = eo3_wo_dataset_doc
    # rewrite lineage to correct format
    doc["lineage"] = {"source_datasets": {"ard": [ls8_eo3_dataset.id]}}
    return doc_to_ds_no_add(
        index,
        wo_eo3_product.name,
        doc, path)


@pytest.fixture
def ga_s2am_ard3_final(index, eo3_sentinel_metadata_type, ga_s2am_ard_3_product, ga_s2am_ard_3_final_doc):
    return doc_to_ds_no_add(
        index,
        ga_s2am_ard_3_product.name,
        *ga_s2am_ard_3_final_doc)


@pytest.fixture
def ga_s2am_ard3_interim(index, eo3_sentinel_metadata_type, ga_s2am_ard_3_product, ga_s2am_ard_3_interim_doc):
    return doc_to_ds_no_add(
        index,
        ga_s2am_ard_3_product.name,
        *ga_s2am_ard_3_interim_doc)


@pytest.fixture
def mem_index_fresh(in_memory_config):
    from datacube import Datacube
    with Datacube(env=in_memory_config) as dc:
        yield dc


@pytest.fixture
def mem_index_eo3(mem_index_fresh,
                  extended_eo3_metadata_type_doc,
                  extended_eo3_product_doc,
                  base_eo3_product_doc):
    mem_index_fresh.index.metadata_types.add(
        mem_index_fresh.index.metadata_types.from_doc(extended_eo3_metadata_type_doc)
    )
    mem_index_fresh.index.products.add_document(base_eo3_product_doc)
    mem_index_fresh.index.products.add_document(extended_eo3_product_doc)
    return mem_index_fresh


@pytest.fixture
def mem_eo3_data(mem_index_eo3, datasets_with_unembedded_lineage_doc):
    (doc_ls8, loc_ls8), (doc_wo, loc_wo) = datasets_with_unembedded_lineage_doc
    from datacube.index.hl import Doc2Dataset
    resolver = Doc2Dataset(mem_index_eo3.index)
    ds_ls8, err = resolver(doc_ls8, loc_ls8)
    mem_index_eo3.index.datasets.add(ds_ls8)
    ds_wo, err = resolver(doc_wo, loc_wo)
    mem_index_eo3.index.datasets.add(ds_wo)
    return mem_index_eo3, ds_ls8.id, ds_wo.id


@pytest.fixture(scope="module", params=["datacube", "experimental"])
def datacube_env_name(request):
    return request.param


@pytest.fixture(params=[("datacube", "experimental"), ("experimental", "datacube")])
def datacube_env_name_pair(request) -> tuple[str, str]:
    return request.param


@pytest.fixture
def odc_config() -> ODCConfig:
    return ODCConfig(paths=CONFIG_FILE_PATHS)


@pytest.fixture
def cfg_env(odc_config, datacube_env_name: str) -> ODCEnvironment:
    """Provides a :class:`ODCEnvironment` configured with suitable config file paths."""
    return odc_config[datacube_env_name]


@pytest.fixture
def cfg_env_pair(odc_config: ODCConfig, datacube_env_name_pair: tuple[str, str]
                 ) -> tuple[ODCEnvironment, ODCEnvironment]:
    """Provides a pair of :class:`ODCEnvironment` configured with suitable config file paths.
    """
    return tuple(odc_config[env] for env in datacube_env_name_pair)


@pytest.fixture
def null_config(odc_config: ODCConfig) -> ODCEnvironment:
    """Provides a :class:`ODCEnvironment` configured with null index driver
    """
    return odc_config["nulldriver"]


@pytest.fixture
def in_memory_config(odc_config: ODCConfig) -> ODCEnvironment:
    """Provides a :class:`ODCEnvironment` configured with memory index driver
    """
    return odc_config["localmemory"]


def reset_db(cfg_env: ODCEnvironment, tz=None) -> PostgresDb | PostGisDb:
    from urllib.parse import urlparse
    url = psql_url_from_config(cfg_env)
    url_components = urlparse(url)
    db_name = url_components.path[1:]
    if cfg_env._name in ('datacube', 'default', 'postgres'):
        db = PostgresDb.from_config(
            cfg_env,
            application_name='test-run',
            validate_connection=False
        )
        # Drop tables so our tests have a clean db.
        # with db.begin() as c:  # Creates a new PostgresDbAPI, by passing a new connection to it
        with db._engine.connect() as connection:
            pgres_core.drop_db(connection)
            if tz:
                connection.execute(text(f'alter database {db_name} set timezone = {tz!r}'))
        # We need to run this as well, I think because SQLAlchemy grabs them into it's MetaData,
        # and attempts to recreate them. WTF TODO FIX
        remove_postgres_dynamic_indexes()
    else:
        db = PostGisDb.from_config(
            cfg_env,
            application_name='test-run',
            validate_connection=False
        )
        with db._engine.connect() as connection:
            pgis_core.drop_db(connection)
            if tz:
                connection.execute(text(f'alter database {db_name} set timezone = {tz!r}'))
        remove_postgis_dynamic_indexes()
    return db


def cleanup_db(cfg_env: ODCEnvironment, db: PostgresDb | PostGisDb):
    with db._engine.connect() as connection:
        if cfg_env._name in ('datacube', 'default', 'postgres'):
            # with db.begin() as c:  # Drop SCHEMA
            pgres_core.drop_db(connection)
        else:
            pgis_core.drop_db(connection)
    db.close()


@pytest.fixture(params=["America/Los_Angeles", "UTC"])
def uninitialised_postgres_db(cfg_env: ODCEnvironment, request) -> PostgresDb | PostGisDb:
    """
    Return a connection to an empty PostgreSQL or PostGIS database
    """
    # Setup
    timezone = request.param
    db = reset_db(cfg_env, timezone)

    yield db

    # Cleanup
    cleanup_db(cfg_env, db)


@pytest.fixture
def uninitialised_postgres_db_pair(cfg_env_pair):
    """
    Return a pair connections to empty PostgreSQL or PostGIS databases
    """
    dbs = tuple(reset_db(cfg_env) for cfg_env in cfg_env_pair)

    yield dbs

    for local_cfg, db in zip(cfg_env_pair, dbs):
        cleanup_db(local_cfg, db)


@pytest.fixture
def index(cfg_env,
          uninitialised_postgres_db: Union[PostGisDb, PostgresDb]):
    index = index_connect(cfg_env, validate_connection=False)
    index.init_db()
    yield index
    del index


@pytest.fixture
def index_pair_populated_empty(cfg_env_pair, uninitialised_postgres_db_pair,
                               extended_eo3_metadata_type_doc,
                               base_eo3_product_doc, extended_eo3_product_doc, africa_s2_product_doc,
                               eo3_ls8_dataset_doc, eo3_ls8_dataset2_doc,
                               eo3_ls8_dataset3_doc, eo3_ls8_dataset4_doc,
                               eo3_wo_dataset_doc, eo3_africa_dataset_doc, eo3_africa_dataset2_doc):
    populated_cfg, empty_cfg = cfg_env_pair
    populated_idx = index_connect(populated_cfg, validate_connection=False)
    empty_idx = index_connect(empty_cfg, validate_connection=False)
    populated_idx.init_db()
    empty_idx.init_db(with_default_types=False)
    assert list(empty_idx.products.get_all()) == []
    assert list(populated_idx.products.get_all()) == []
    # Populate the populated index
    populated_idx.metadata_types.add(
        populated_idx.metadata_types.from_doc(extended_eo3_metadata_type_doc)
    )
    for prod_doc in (base_eo3_product_doc, extended_eo3_product_doc, africa_s2_product_doc):
        populated_idx.products.add_document(prod_doc)
    for ds_doc, ds_path in (eo3_ls8_dataset_doc, eo3_ls8_dataset2_doc,
                            eo3_ls8_dataset3_doc, eo3_ls8_dataset4_doc,
                            eo3_wo_dataset_doc, eo3_africa_dataset_doc, eo3_africa_dataset2_doc):
        doc_to_ds(populated_idx, ds_doc['product']['name'], ds_doc, ds_path)
    assert list(populated_idx.products.get_all()) != list(empty_idx.products.get_all())
    assert list(empty_idx.products.get_all()) == []
    assert list(populated_idx.products.get_all()) != []

    yield (populated_idx, empty_idx)

    del populated_idx
    del empty_idx


@pytest.fixture
def index_empty(cfg_env, uninitialised_postgres_db: Union[PostGisDb, PostgresDb]):
    index = index_connect(cfg_env, validate_connection=False)
    index.init_db(with_default_types=False)
    yield index
    del index


def remove_postgres_dynamic_indexes():
    """
    Clear any dynamically created postgresql indexes from the schema.
    """
    # Our normal indexes start with "ix_", dynamic indexes with "dix_"
    for table in pgres_core.METADATA.tables.values():
        table.indexes.intersection_update([i for i in table.indexes if not i.name.startswith('dix_')])


def remove_postgis_dynamic_indexes():
    """
    Clear any dynamically created postgis indexes from the schema.
    """
    # Our normal indexes start with "ix_", dynamic indexes with "dix_"
    # for table in pgis_core.METADATA.tables.values():
    #    table.indexes.intersection_update([i for i in table.indexes if not i.name.startswith('dix_')])
    # Dynamic indexes disabled.


@pytest.fixture
def ls5_telem_doc(ga_metadata_type):
    return {
        "name": "ls5_telem_test",
        "description": 'LS5 Test',
        "license": "CC-BY-4.0",
        "metadata": {
            "platform": {
                "code": "LANDSAT_5"
            },
            "product_type": "satellite_telemetry_data",
            "ga_level": "P00",
            "format": {
                "name": "RCC"
            }
        },
        "metadata_type": ga_metadata_type.name
    }


@pytest.fixture
def ls5_telem_type(index, ls5_telem_doc):
    return index.products.add_document(ls5_telem_doc)


@pytest.fixture(scope='session')
def geotiffs(tmpdir_factory):
    """Create test geotiffs and corresponding yamls.

    We create one yaml per time slice, itself comprising one geotiff
    per band, each with specific custom data that can be later
    tested. These are meant to be used by all tests in the current
    session, by way of symlinking the yamls and tiffs returned by this
    fixture, in order to save disk space (and potentially generation
    time).

    The yamls are customised versions of
    :ref:`_EXAMPLE_LS5_NBAR_DATASET_FILE` shifted by 24h and with
    spatial coords reflecting the size of the test geotiff, defined in
    :ref:`GEOTIFF`.

    :param tmpdir_fatory: pytest tmp dir factory.
    :return: List of dictionaries like::

        {
            'day':..., # compact day string, e.g. `19900302`
            'uuid':..., # a unique UUID for this dataset (i.e. specific day)
            'path':..., # path to the yaml ingestion file
            'tiffs':... # list of paths to the actual geotiffs in that dataset, one per band.
        }

    """
    tiffs_dir = tmpdir_factory.mktemp('tiffs')

    config = load_yaml_file(_EXAMPLE_LS5_NBAR_DATASET_FILE)[0]

    # Customise the spatial coordinates
    ul = GEOTIFF['ul']
    lr = {
        dim: ul[dim] + GEOTIFF['shape'][dim] * GEOTIFF['pixel_size'][dim]
        for dim in ('x', 'y')
    }
    config['grid_spatial']['projection']['geo_ref_points'] = {
        'ul': ul,
        'ur': {
            'x': lr['x'],
            'y': ul['y']
        },
        'll': {
            'x': ul['x'],
            'y': lr['y']
        },
        'lr': lr
    }
    # Generate the custom geotiff yamls
    return [_make_tiffs_and_yamls(tiffs_dir, config, day_offset)
            for day_offset in range(NUM_TIME_SLICES)]


def _make_tiffs_and_yamls(tiffs_dir, config, day_offset):
    """Make a custom yaml and tiff for a day offset.

    :param path-like tiffs_dir: The base path to receive the tiffs.
    :param dict config: The yaml config to be cloned and altered.
    :param int day_offset: how many days to offset the original yaml by.
    """
    config = deepcopy(config)

    # Increment all dates by the day_offset
    delta = timedelta(days=day_offset)
    day_orig = config['acquisition']['aos'].strftime('%Y%m%d')
    config['acquisition']['aos'] += delta
    config['acquisition']['los'] += delta
    config['extent']['from_dt'] += delta
    config['extent']['center_dt'] += delta
    config['extent']['to_dt'] += delta
    day = config['acquisition']['aos'].strftime('%Y%m%d')

    # Set the main UUID and assign random UUIDs where needed
    uuid = uuid4()
    config['id'] = str(uuid)
    level1 = config['lineage']['source_datasets']['level1']
    level1['id'] = str(uuid4())
    level1['lineage']['source_datasets']['satellite_telemetry_data']['id'] = str(uuid4())

    # Alter band data
    bands = config['image']['bands']
    for band in bands.keys():
        # Copy dict to avoid aliases in yaml output (for better legibility)
        bands[band]['shape'] = copy(GEOTIFF['shape'])
        bands[band]['cell_size'] = {
            dim: abs(GEOTIFF['pixel_size'][dim]) for dim in ('x', 'y')}
        bands[band]['path'] = bands[band]['path'].replace('product/', '').replace(day_orig, day)

    dest_path = str(tiffs_dir.join('agdc-metadata_%s.yaml' % day))
    with open(dest_path, 'w') as dest_yaml:
        yaml.dump(config, dest_yaml)
    return {
        'day': day,
        'uuid': uuid,
        'path': dest_path,
        'tiffs': _make_geotiffs(tiffs_dir, day_offset)  # make 1 geotiff per band
    }


@pytest.fixture
def example_ls5_dataset_path(example_ls5_dataset_paths):
    """Create a single sample raw observation (dataset + geotiff)."""
    return list(example_ls5_dataset_paths.values())[0]


@pytest.fixture
def example_ls5_dataset_paths(tmpdir, geotiffs):
    """Create sample raw observations (dataset + geotiff).

    This fixture should be used by eah test requiring a set of
    observations over multiple time slices. The actual geotiffs and
    corresponding yamls are symlinks to a set created for the whole
    test session, in order to save disk and time.

    :param tmpdir: The temp directory in which to create the datasets.
    :param list geotiffs: List of session geotiffs and yamls, to be
      linked from this unique observation set sample.
    :return: dict: Dict of directories containing each observation,
      indexed by dataset UUID.
    """
    dataset_dirs = _make_ls5_scene_datasets(geotiffs, tmpdir)
    return dataset_dirs


@pytest.fixture
def default_metadata_type_doc():
    return [doc for doc in default_metadata_type_docs() if doc['name'] == 'eo'][0]


@pytest.fixture
def eo3_metadata_type_docs(eo3_base_metadata_type_doc, extended_eo3_metadata_type_doc):
    return [eo3_base_metadata_type_doc, extended_eo3_metadata_type_doc]


@pytest.fixture
def eo3_base_metadata_type_doc():
    return [doc for doc in default_metadata_type_docs() if doc['name'] == 'eo3'][0]


@pytest.fixture
def telemetry_metadata_type_doc():
    return [doc for doc in default_metadata_type_docs() if doc['name'] == 'telemetry'][0]


@pytest.fixture
def ga_metadata_type_doc():
    _FULL_EO_METADATA = Path(__file__).parent.joinpath('extensive-eo-metadata.yaml')  # noqa: N806
    [(path, eo_md_type)] = datacube.utils.read_documents(_FULL_EO_METADATA)
    return eo_md_type


@pytest.fixture
def default_metadata_types(index, eo3_metadata_type_docs):
    """Inserts the default metadata types into the Index"""
    if index.supports_legacy:
        type_docs = default_metadata_type_docs()
    else:
        type_docs = eo3_metadata_type_docs
    for d in type_docs:
        index.metadata_types.add(index.metadata_types.from_doc(d))
    return index.metadata_types.get_all()


@pytest.fixture
def ga_metadata_type(index, ga_metadata_type_doc):
    return index.metadata_types.add(index.metadata_types.from_doc(ga_metadata_type_doc))


@pytest.fixture
def default_metadata_type(index, default_metadata_types):
    if index.supports_legacy:
        return index.metadata_types.get_by_name('eo')
    else:
        return index.metadata_types.get_by_name('eo3')


@pytest.fixture
def telemetry_metadata_type(index, default_metadata_types):
    return index.metadata_types.get_by_name('telemetry')


@pytest.fixture
def indexed_ls5_scene_products(index, ga_metadata_type):
    """Add Landsat 5 scene Products into the Index"""
    products = load_test_products(
        CONFIG_SAMPLES / 'dataset_types' / 'ls5_scenes.yaml',
        # Use our larger metadata type with a more diverse set of field types.
        metadata_type=ga_metadata_type
    )

    types = []
    for product in products:
        types.append(index.products.add_document(product))

    return types


@pytest.fixture
def example_ls5_nbar_metadata_doc():
    return load_yaml_file(_EXAMPLE_LS5_NBAR_DATASET_FILE)[0]


@pytest.fixture
def clirunner(datacube_env_name):
    def _run_cli(opts, catch_exceptions=False,
                 expect_success=True, cli_method=datacube.scripts.cli_app.cli,
                 skip_env=False, skip_config_paths=False,
                 mix_stderr=True,
                 verbose_flag='-v'):
        # If raw config passed in, skip default test config
        if not skip_config_paths:
            exe_opts = list(itertools.chain(*(('--config', f) for f in CONFIG_FILE_PATHS)))
        if not skip_env:
            exe_opts += ['--env', datacube_env_name]
        if verbose_flag:
            exe_opts.append(verbose_flag)
        exe_opts.extend(opts)

        runner = CliRunner(mix_stderr=mix_stderr)
        result = runner.invoke(
            cli_method,
            exe_opts,
            catch_exceptions=catch_exceptions
        )
        if expect_success:
            assert 0 == result.exit_code, "Error for %r. output: %r" % (opts, result.output)
        return result

    return _run_cli


@pytest.fixture
def clirunner_raw():
    def _run_cli(opts,
                 catch_exceptions=False,
                 expect_success=True,
                 cli_method=datacube.scripts.cli_app.cli,
                 verbose_flag='-v'):
        exe_opts = []
        if verbose_flag:
            exe_opts.append(verbose_flag)
        exe_opts.extend(opts)

        runner = CliRunner()
        result = runner.invoke(
            cli_method,
            exe_opts,
            catch_exceptions=catch_exceptions
        )
        if expect_success:
            assert 0 == result.exit_code, "Error for %r. output: %r" % (opts, result.output)
        return result

    return _run_cli


@pytest.fixture
def dataset_add_configs():
    B = INTEGRATION_TESTS_DIR / 'data' / 'dataset_add'
    return SimpleNamespace(metadata=str(B / 'metadata.yml'),
                           products=str(B / 'products.yml'),
                           datasets_bad1=str(B / 'datasets_bad1.yml'),
                           datasets_no_id=str(B / 'datasets_no_id.yml'),
                           datasets_eo3=str(B / 'datasets_eo3.yml'),
                           datasets_eo3_updated=str(B / 'datasets_eo3_updated.yml'),
                           datasets=str(B / 'datasets.yml'),
                           empty_file=str(B / 'empty_file.yml'))


@pytest.fixture
def src_tree_ids():
    return {
        "root": uuid4(),
        "ard1": uuid4(),
        "ard2": uuid4(),

        "l1_1": uuid4(),
        "l1_2": uuid4(),
        "l1_3": uuid4(),

        "l1_4": uuid4(),
        "l1_5": uuid4(),
        "l1_6": uuid4(),

        "atmos": uuid4(),
        "atmos_parent": uuid4()
    }


@pytest.fixture
def src_lineage_tree(src_tree_ids):
    ids = src_tree_ids
    direction = LineageDirection.SOURCES
    return LineageTree(
        dataset_id=ids["root"], direction=direction,
        children={
            "ard": [
                LineageTree(
                    dataset_id=ids["ard1"], direction=direction,
                    children={
                        "l1": [
                            LineageTree(
                                dataset_id=ids["l1_1"], direction=direction,
                                home="level1",
                                children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_2"], direction=direction,
                                home="level1",
                                children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_3"], direction=direction,
                                home="level1",
                                children={}
                            ),
                        ],
                        "atmos_corr": [
                            LineageTree(
                                dataset_id=ids["atmos"], direction=direction,
                                home="anciliary",
                                children=None
                            )
                        ],
                    }
                ),
                LineageTree(
                    dataset_id=ids["ard2"], direction=direction,
                    children={
                        "l1": [
                            LineageTree(
                                dataset_id=ids["l1_4"], direction=direction,
                                home="level1",
                                children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_5"], direction=direction,
                                home="level1",
                                children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_6"], direction=direction,
                                home="level1",
                                children={}
                            ),
                        ],
                        "atmos_corr": [
                            LineageTree(
                                dataset_id=ids["atmos"], direction=direction,
                                home="anciliary",
                                children={
                                    "preatmos": [
                                        LineageTree(
                                            dataset_id=ids["atmos_parent"], direction=direction,
                                            home="anciliary",
                                            children={}
                                        )
                                    ]
                                }
                            )
                        ],
                    }
                ),
            ]
        }
    ), ids


@pytest.fixture
def compatible_derived_tree(src_tree_ids):
    ids = src_tree_ids.copy()
    ids.update({
        "atmos_grandparent": uuid4(),
        "ard3": uuid4(),
        "ard4": uuid4(),
        "leaf_1": uuid4(),
        "leaf_2": uuid4(),
        "leaf_3": uuid4(),
        "child_of_root": uuid4(),
        "grandchild_of_root": uuid4(),
    })
    tree = LineageTree(
        dataset_id=ids["atmos_grandparent"],
        direction=LineageDirection.DERIVED,
        home="steves_basement",
        children={
            "spam": [
                LineageTree(
                    dataset_id=ids["atmos_parent"],
                    direction=LineageDirection.DERIVED,
                    home="anciliary",
                    children={
                        "preatmos": [
                            LineageTree(
                                dataset_id=ids["atmos"],
                                direction=LineageDirection.DERIVED,
                                home="anciliary",
                                children={
                                    "atmos_corr": [
                                        LineageTree(
                                            dataset_id=ids["ard1"],
                                            direction=LineageDirection.DERIVED,
                                            home="ard",
                                            children={
                                                "ard": [
                                                    LineageTree(
                                                        dataset_id=ids["root"],
                                                        direction=LineageDirection.DERIVED,
                                                        home="extensions", children={
                                                            "dra": [
                                                                LineageTree(
                                                                    dataset_id=ids["child_of_root"],
                                                                    direction=LineageDirection.DERIVED,
                                                                    home="extensions", children={
                                                                        "rad": [
                                                                            LineageTree(
                                                                                dataset_id=ids["grandchild_of_root"],
                                                                                direction=LineageDirection.DERIVED,
                                                                                home="extensions", children={}
                                                                            )

                                                                        ]
                                                                    }
                                                                )
                                                            ]
                                                        }
                                                    ),
                                                    LineageTree(
                                                        dataset_id=ids["leaf_1"],
                                                        direction=LineageDirection.DERIVED,
                                                        home="extensions", children={}
                                                    ),
                                                ]
                                            }
                                        ),
                                        LineageTree(
                                            dataset_id=ids["ard2"],
                                            direction=LineageDirection.DERIVED,
                                            home="ard",
                                            children={}
                                        ),
                                        LineageTree(
                                            dataset_id=ids["ard3"],
                                            direction=LineageDirection.DERIVED,
                                            home="ard",
                                            children={
                                                "ard": [
                                                    LineageTree(
                                                        dataset_id=ids["leaf_2"],
                                                        direction=LineageDirection.DERIVED,
                                                        home="extensions", children={}
                                                    ),
                                                    LineageTree(
                                                        dataset_id=ids["leaf_3"],
                                                        direction=LineageDirection.DERIVED,
                                                        home="extensions", children={}
                                                    ),
                                                ]
                                            }
                                        ),
                                        LineageTree(
                                            dataset_id=ids["ard4"],
                                            direction=LineageDirection.DERIVED,
                                            home="ard",
                                            children={}
                                        ),
                                    ]
                                }
                            )
                        ]
                    }
                )
            ]
        }
    )
    return tree, ids


@pytest.fixture
def dataset_with_external_lineage(index,
                                  src_lineage_tree, compatible_derived_tree,
                                  ls8_eo3_product, eo3_ls8_dataset_doc):
    src_tree, ids = src_lineage_tree
    derived_tree, ids = compatible_derived_tree
    eo3_ls8_dataset_doc[0]["id"] = ids["root"]
    dataset = doc_to_ds(index, ls8_eo3_product.name, *eo3_ls8_dataset_doc,
                        src_tree=src_tree, derived_tree=derived_tree)
    return dataset, src_tree, derived_tree, ids



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from typing import NamedTuple, Iterable, cast, Sequence
from uuid import UUID

from deprecat import deprecat

from datacube.migration import ODC2DeprecationWarning
from datacube.model import Product, Dataset
from datacube.utils import cached_property
from datacube.utils.documents import JsonDict


class BatchStatus(NamedTuple):
    """
    A named tuple representing the results of a batch add operation:
    - completed: Number of objects added to theMay be None for internal functions and for datasets.
    - skipped: Number of objects skipped, either because they already exist
      or the documents are invalid for this driver.
    - seconds_elapsed: seconds elapsed during the bulk add operation;
    - safe: an optional list of names of bulk added objects that are safe to be
      used for lower level bulk adds. Includes objects added, and objects skipped
      because they already exist in the index and are identical to the version
      being added.  May be None for internal functions and for datasets.
    """
    completed: int
    skipped: int
    seconds_elapsed: float
    safe: Iterable[str] | None = None


# Non-strict Dataset ID representation

DSID = str | UUID


def dsid_to_uuid(dsid: DSID) -> UUID:
    """
    Convert non-strict dataset ID representation to strict UUID
    """
    if isinstance(dsid, UUID):
        return dsid
    else:
        return UUID(dsid)


class DatasetTuple(NamedTuple):
    """
    A named tuple representing a complete dataset:
    - product: A Product model.
    - metadata: The dataset metadata document
    - uri_: The dataset location or list of locations
    """
    product: Product
    metadata: JsonDict
    uri_: str | list[str]

    @property
    def uri_is_string(self):
        if isinstance(self.uri_, str):
            return True
        return False

    @property
    def is_legacy(self):
        if self.uri_is_string:
            return False
        return len(self.uri_) > 1

    @property
    def uri(self) -> str:
        if self.uri_is_string:
            return cast(str, self.uri_)
        elif self.is_legacy:
            return self.uris[0]
        else:
            return cast(list[str], self.uri_)[0]

    @property
    @deprecat(
        reason="Multiple uris are deprecated. Please use the uri field and ensure that datasets only have one location",
        version='1.9.0',
        category=ODC2DeprecationWarning)
    def uris(self) -> Sequence[str]:
        if self.uri_is_string:
            return [cast(str, self.uri_)]
        else:
            return cast(list[str], self.uri_)


# The special handling of grid_spatial, etc appears to NOT apply to EO3.
# Does EO3 handle it in metadata?
class DatasetSpatialMixin:
    __slots__ = ()

    @property
    def _gs(self):
        return self.grid_spatial

    @property
    def crs(self):
        return Dataset.crs.__get__(self)

    @cached_property
    def extent(self):
        return Dataset.extent.func(self)

    @property
    def transform(self):
        return Dataset.transform.__get__(self)

    @property
    def bounds(self):
        return Dataset.bounds.__get__(self)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import pytest

from datacube.index.postgres.index import Index


@pytest.mark.parametrize('datacube_env_name', ('datacube', ))
def test_with_standard_index(uninitialised_postgres_db, cfg_env):
    index = Index(uninitialised_postgres_db, cfg_env)
    index.init_db()


def test_system_init(uninitialised_postgres_db, clirunner):
    result = clirunner(['system', 'init'], catch_exceptions=False)

    # Question: Should the Index be able to be specified on the command line, or should it come from the config file?

    if result.exit_code != 0:
        print(result.output)
        assert False



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from time import monotonic
from typing import Iterable

from datacube.index.abstract import BatchStatus, NoLineageResource
from datacube.index.postgres._transaction import IndexResourceAddIn
from datacube.model import LineageRelation


class LineageResource(NoLineageResource, IndexResourceAddIn):
    def __init__(self, db, index):
        self._db = db
        super().__init__(index)

    def get_all_lineage(self, batch_size: int = 1000) -> Iterable[LineageRelation]:
        with self._db_connection(transaction=True) as connection:
            for row in connection.get_all_lineage(batch_size=batch_size):
                yield LineageRelation(
                    derived_id=row.dataset_ref,
                    classifier=row.classifier,
                    source_id=row.source_dataset_ref
                )

    def _add_batch(self, batch: Iterable[LineageRelation]) -> BatchStatus:
        b_started = monotonic()
        with self._db_connection(transaction=True) as connection:
            b_added, b_skipped = connection.insert_lineage_bulk(
                [
                    (str(rel.derived_id), rel.classifier, str(rel.source_id))
                    for rel in batch
                ]
            )
        return BatchStatus(b_added, b_skipped, monotonic() - b_started)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0

import pytest
from odc.geo.geom import box, CRS


def test_extract_geom():
    p1 = box(-122.2, 44.7, -120.8, 45.0, crs="epsg:4326")
    p2 = box(-102.5, 44.1, -100.2, 46.2, crs="epsg:4326")
    p3 = box(-112.7, 42.4, -109.6, 45.1, crs="epsg:4326")

    from datacube.index._spatial import extract_geom_from_query
    geom = extract_geom_from_query(geopolygon=[p1, p2, p3])
    assert geom.contains(p1)
    assert geom.contains(p2)
    assert geom.contains(p3)
    geom = extract_geom_from_query(lon=(-122.3, -100.2))
    assert geom.crs == CRS("epsg:4326")
    assert geom.contains(p1)
    assert geom.contains(p2)
    assert geom.contains(p3)

    with pytest.raises(ValueError):
        geom = extract_geom_from_query(geopolygon=p3, crs="epsg:3577")

    with pytest.raises(ValueError):
        geom = extract_geom_from_query(latitude=(22, 23), y=(10011.1, 585585.5), crs="epsg:3577")

    with pytest.raises(ValueError):
        geom = extract_geom_from_query(lon=(22, 23), x=(10011.1, 585585.5), crs="epsg:3577")



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0


def test_spindex_table_name():
    from datacube.drivers.postgis._spatial import is_spindex_table_name
    assert is_spindex_table_name("spatial_4326")
    assert not is_spindex_table_name("spatial")
    assert not is_spindex_table_name("spatial_")
    assert not is_spindex_table_name("spatial_0")
    assert not is_spindex_table_name("spam_spam")
    assert not is_spindex_table_name("spatial_spam")
    assert not is_spindex_table_name("spatial_-4326")
    assert not is_spindex_table_name("spatial_4326_spam")
    assert not is_spindex_table_name("spatial_spam_4326")



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import numpy.testing as npt
import pytest
import xarray as xr
import pandas as pd
import numpy as np
from datacube.testutils.geom import epsg4326, epsg3857
from datacube.testutils import mk_sample_xr_dataset, remove_crs
from datacube.utils.xarray_geoextensions import _xarray_affine

multi_coords = xr.DataArray(
    np.zeros(1),
    [
        (
            "spec",
            pd.MultiIndex.from_arrays(
                np.array([["2001-01-01"], ["2001-01-01"]]), names=("time", "solar_day")
            ),
        )
    ],
).coords
single_coord = dict(time=np.array(["2001-01-01"]))


@pytest.mark.parametrize(
    "odc_style_xr_dataset", [single_coord, multi_coords], indirect=True
)
def test_xr_extension(odc_style_xr_dataset):
    xx = odc_style_xr_dataset

    assert (1,) + xx.odc.geobox.shape == xx.B10.shape

    (sx, zz0, tx, zz1, sy, ty) = xx.affine[:6]
    assert (zz0, zz1) == (0, 0)

    xx = remove_crs(xx)

    # affine should still be valid
    A = _xarray_affine(xx)
    assert A is not None
    npt.assert_allclose(A * (0.5, 0.5), (xx.longitude.data[0], xx.latitude.data[0]))
    npt.assert_allclose(A * (0.5, 1.5), (xx.longitude.data[0], xx.latitude.data[1]))


def test_xr_geobox():
    xy = (10, 111)
    rxy = (10, -100)
    resolution = rxy[::-1]

    ds = mk_sample_xr_dataset(crs=epsg3857, xy=xy, resolution=resolution)

    assert ds.odc.geobox.crs == epsg3857
    assert ds.band.odc.geobox.crs == epsg3857
    assert ds.band.affine * (0, 0) == xy
    assert ds.band.affine * (1, 1) == tuple(a + b for a, b in zip(xy, rxy))

    assert ds.band[:, :2, :2].affine * (0, 0) == xy
    assert ds.band[:, :2, :2].affine * (1, 1) == tuple(a + b for a, b in zip(xy, rxy))
    assert ds.band.isel(time=0, x=0).affine is None

    xx = ds.band + 1000
    assert xx.odc.geobox is not None
    assert xx.odc.geobox == ds.band.odc.geobox

    assert mk_sample_xr_dataset(crs=epsg4326).odc.geobox.crs == epsg4326
    assert mk_sample_xr_dataset(crs=epsg4326).band.odc.geobox.crs == epsg4326

    assert mk_sample_xr_dataset(crs=None).affine is not None



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import datetime
import pytest

from datacube.model import Range
from datacube.index import Index
from odc.geo import CRS


@pytest.mark.parametrize('datacube_env_name', ('experimental',))
def test_index_environment(index: Index):
    assert index.environment.index_driver in ("experimental", "postgis")


@pytest.mark.parametrize('datacube_env_name', ('experimental',))
def test_create_drop_spatial_index(index: Index):
    # Default spatial index for 4326
    assert list(index.spatial_indexes()) == [CRS("epsg:4326")]
    # WKT CRS which cannot be mapped to an EPSG number.
    assert not index.create_spatial_index(CRS(
        'GEOGCS["GCS_WGS_1984",DATUM["D_WGS_1984",SPHEROID["WGS_1984",6378137,298.257223563]]'
        ',PRIMEM["Weird",22.3],UNIT["Degree",0.017453292519943295]]'
    ))
    assert list(index.spatial_indexes()) == [CRS("epsg:4326")]
    assert index.create_spatial_index(CRS("epsg:3577"))
    assert index.create_spatial_index(CRS("WGS-84"))
    assert set(index.spatial_indexes()) == {CRS("epsg:3577"), CRS("epsg:4326")}
    assert set(index.spatial_indexes(refresh=True)) == {CRS("epsg:3577"), CRS("epsg:4326")}
    assert index.drop_spatial_index(CRS("epsg:3577"))
    assert index.spatial_indexes() == [CRS("epsg:4326")]
    assert index.spatial_indexes(refresh=True) == [CRS("epsg:4326")]


@pytest.mark.parametrize('datacube_env_name', ('experimental',))
def test_spatial_index_maintain(index: Index, ls8_eo3_product, eo3_ls8_dataset_doc):
    index.create_spatial_index(CRS("EPSG:3577"))
    assert set(index.spatial_indexes(refresh=True)) == {CRS("EPSG:3577"), CRS("EPSG:4326")}
    from datacube.index.hl import Doc2Dataset
    resolver = Doc2Dataset(index, products=[ls8_eo3_product.name], verify_lineage=False)
    ds, err = resolver(*eo3_ls8_dataset_doc)
    assert err is None and ds is not None
    ds = index.datasets.add(ds, False)
    assert ds
    index.datasets.archive([ds.id])
    index.datasets.purge([ds.id])
    # Can't really read yet, but seems to write at least


@pytest.mark.parametrize('datacube_env_name', ('experimental',))
def test_spatial_index_populate(index: Index,
                                ls8_eo3_product,
                                wo_eo3_product,
                                ls8_eo3_dataset, ls8_eo3_dataset2,
                                ls8_eo3_dataset3, ls8_eo3_dataset4,
                                wo_eo3_dataset):
    index.create_spatial_index(CRS("EPSG:3577"))
    assert set(index.spatial_indexes(refresh=True)) == {CRS("EPSG:3577"), CRS("EPSG:4326")}
    assert index.update_spatial_index(
        crses=[CRS("EPSG:4326")],
        dataset_ids=[ls8_eo3_dataset.id, ls8_eo3_dataset2.id]
    ) == 2
    assert index.update_spatial_index(product_names=[ls8_eo3_product.name]) == 8
    assert index.update_spatial_index() == 10
    assert index.update_spatial_index(
        crses=[CRS("EPSG:4326")],
        product_names=[wo_eo3_product.name],
        dataset_ids=[ls8_eo3_dataset.id]
    ) == 2
    assert index.update_spatial_index(product_names=[ls8_eo3_product.name], dataset_ids=[ls8_eo3_dataset.id]) == 8


@pytest.mark.parametrize('datacube_env_name', ('experimental',))
def test_spatial_index_crs_validity(index: Index,
                                    ls8_eo3_product, ls8_eo3_dataset,
                                    africa_s2_eo3_product, africa_eo3_dataset):
    # TODO: potentially include africa_eo3_dataset2 in this test
    epsg4326 = CRS("EPSG:4326")
    epsg3577 = CRS("EPSG:3577")
    index.create_spatial_index(epsg3577)
    assert set(index.spatial_indexes(refresh=True)) == {epsg4326, epsg3577}
    assert index.update_spatial_index(crses=[epsg3577]) == 2


def test_spatial_index_crs_sanitise():
    epsg4326 = CRS("EPSG:4326")
    epsg3857 = CRS("EPSG:3857")
    epsg3832 = CRS("EPSG:3832")
    from odc.geo.geom import polygon

    # 4326, crossing the Prime-Meridian, wound clockwise
    pm_4326_cw = polygon((
        (-2.0, 25.0),
        (2.0, 25.0),
        (2.0, 23.0),
        (-2.0, 23.0),
        (-2.0, 25.0),
    ), crs=epsg4326)

    # 4326, crossing the Prime-Meridian, wound counter-clockwise
    pm_4326_ccw = polygon((
        (-2.0, 25.0),
        (-2.0, 23.0),
        (2.0, 23.0),
        (2.0, 25.0),
        (-2.0, 25.0),
    ), crs=epsg4326)

    # 4326, crossing the Prime-Meridian, wound clockwise
    am_4326_ccw = polygon((
        (178.0, 25.0),
        (178.0, 23.0),
        (-178.0, 23.0),
        (-178.0, 25.0),
        (178.0, 25.0),
    ), crs=epsg4326)

    # 4326, crossing the Prime-Meridian, wound counter-clockwise
    am_4326_cw = polygon((
        (178.0, 25.0),
        (-178.0, 25.0),
        (-178.0, 23.0),
        (178.0, 23.0),
        (178.0, 25.0),
    ), crs=epsg4326)

    pm_3857 = polygon((
        (-222638, 2875744),
        (-222638, 2632018),
        (222638, 2632018),
        (222638, 2632018),
        (-222638, 2875744),
    ), crs=epsg3857)

    am_3857 = polygon((
        (-19814869, 2875744),


        (19814869, 2632018),


        (-19814869, 2875744),
    ), crs=epsg3857)

    am_3832 = polygon((
        (3116945, 2857692),
        (3562223, 2857692),
        (3562223, 2615329),
        (3116945, 2615329),
        (3116945, 2857692)
    ), crs=epsg3832)

    from datacube.drivers.postgis._spatial import sanitise_extent

    assert sanitise_extent(pm_4326_ccw, epsg4326) == pm_4326_ccw
    assert sanitise_extent(pm_4326_cw, epsg4326) == pm_4326_ccw
    assert sanitise_extent(pm_4326_ccw, epsg3857) == pm_4326_ccw.to_crs(epsg3857)

    sanitised = sanitise_extent(am_4326_ccw, epsg4326)
    assert sanitised.type == "MultiPolygon"
    assert sanitised.json["coordinates"] == [
        ((
            (180.0, 25.0),
            (178.0, 25.0),
            (178.0, 23.0),
            (180.0, 23.0),
            (180.0, 25.0),
        ),),
        ((
            (-180.0, 23.0),
            (-178.0, 23.0),
            (-178.0, 25.0),
            (-180.0, 25.0),
            (-180.0, 23.0),
        ),),
    ]

    assert sanitise_extent(pm_4326_ccw, epsg3857) == pm_4326_ccw.to_crs(epsg3857)
    assert sanitise_extent(pm_3857, epsg3857).geom.almost_equals(pm_3857.geom)
    assert sanitise_extent(am_3857, epsg3857).type == "MultiPolygon"

    assert sanitise_extent(am_3832, epsg3832) == am_3832
    assert sanitise_extent(am_3832, epsg4326).type == "MultiPolygon"


@pytest.mark.parametrize('datacube_env_name', ('experimental',))
def test_spatial_extent(index,
                        ls8_eo3_dataset, ls8_eo3_dataset2,
                        ls8_eo3_dataset3, ls8_eo3_dataset4,
                        africa_s2_eo3_product, africa_eo3_dataset):
    # TODO: include africa_eo3_dataset2 in this test
    epsg4326 = CRS("EPSG:4326")
    epsg3577 = CRS("EPSG:3577")
    index.create_spatial_index(epsg3577)
    index.update_spatial_index(crses=[epsg3577])

    with pytest.raises(KeyError):
        index.products.spatial_extent("spaghetti_product")

    ext1 = index.datasets.spatial_extent([ls8_eo3_dataset.id], crs=epsg4326)
    ext2 = index.datasets.spatial_extent([ls8_eo3_dataset2.id], crs=epsg4326)
    ext12 = index.datasets.spatial_extent([ls8_eo3_dataset.id, ls8_eo3_dataset2.id], crs=epsg4326)
    assert ext1 is not None and ext2 is not None and ext12 is not None
    assert ext1 == ext2
    assert ext12.difference(ext1).area < 0.001
    assert ls8_eo3_dataset.extent.to_crs(epsg4326).intersects(ext1)
    assert ls8_eo3_dataset.extent.to_crs(epsg4326).intersects(ext12)
    assert ls8_eo3_dataset2.extent.to_crs(epsg4326).intersects(ext2)
    assert ls8_eo3_dataset2.extent.to_crs(epsg4326).intersects(ext12)
    extau12 = index.datasets.spatial_extent([ls8_eo3_dataset.id, ls8_eo3_dataset2.id], crs=epsg3577)
    extau12africa = index.datasets.spatial_extent(
        [ls8_eo3_dataset.id, ls8_eo3_dataset2.id, africa_eo3_dataset.id],
        crs=epsg3577
    )
    assert extau12.area < extau12africa.area
    ext3 = index.datasets.spatial_extent(ids=[ls8_eo3_dataset3.id], crs=epsg4326)
    ext1234 = index.datasets.spatial_extent(
        [
            ls8_eo3_dataset.id, ls8_eo3_dataset2.id,
            ls8_eo3_dataset3.id, ls8_eo3_dataset4.id
        ],
        crs=epsg4326)
    assert ext1.difference(ext1234).area < 0.001
    assert ext3.difference(ext1234).area < 0.001
    ext1_3577 = index.datasets.spatial_extent([ls8_eo3_dataset.id], crs=epsg3577)
    assert ext1_3577.intersects(ls8_eo3_dataset.extent._to_crs(epsg3577))

    ext_ls8 = index.products.spatial_extent(
        ls8_eo3_dataset.product,
        crs=epsg4326
    )
    assert ext_ls8 == ext1234
    ext_ls8 = index.products.spatial_extent(
        ls8_eo3_dataset.product.name,
        crs=epsg4326
    )
    assert ext_ls8 == ext1234


@pytest.mark.parametrize('datacube_env_name', ('experimental',))
def test_spatial_search(index,
                        ls8_eo3_dataset, ls8_eo3_dataset2,
                        ls8_eo3_dataset3, ls8_eo3_dataset4):
    epsg4326 = CRS("EPSG:4326")
    epsg3577 = CRS("EPSG:3577")
    index.create_spatial_index(epsg3577)
    index.update_spatial_index(crses=[epsg3577])
    # Test old style lat/lon search
    dss = index.datasets.search(
        product=ls8_eo3_dataset.product.name,
        lat=Range(begin=-37.5, end=37.0),
        lon=Range(begin=148.5, end=149.0)
    )
    dssids = [ds.id for ds in dss]
    assert len(dssids) == 2
    assert ls8_eo3_dataset.id in dssids
    assert ls8_eo3_dataset2.id in dssids
    # Test polygons
    exact1_4326 = ls8_eo3_dataset.extent.to_crs(epsg4326)
    exact1_3577 = ls8_eo3_dataset.extent.to_crs(epsg3577)
    exact3_4326 = ls8_eo3_dataset3.extent.to_crs(epsg4326)
    exact3_3577 = ls8_eo3_dataset3.extent.to_crs(epsg3577)
    dssids = set(ds.id for ds in index.datasets.search(product=ls8_eo3_dataset.product.name, geopolygon=exact1_4326))
    assert len(dssids) == 2
    assert ls8_eo3_dataset.id in dssids
    assert ls8_eo3_dataset2.id in dssids
    assert index.datasets.count(product=ls8_eo3_dataset.product.name, geopolygon=exact1_4326) == 2
    dssids = [ds.id for ds in index.datasets.search(product=ls8_eo3_dataset.product.name, geopolygon=exact1_3577)]
    assert len(dssids) == 2
    assert ls8_eo3_dataset.id in dssids
    assert ls8_eo3_dataset2.id in dssids
    assert index.datasets.count(product=ls8_eo3_dataset.product.name, geopolygon=exact1_3577) == 2
    dssids = [ds.id for ds in index.datasets.search(product=ls8_eo3_dataset.product.name, geopolygon=exact3_4326)]
    assert len(dssids) == 2
    assert ls8_eo3_dataset3.id in dssids
    assert ls8_eo3_dataset3.id in dssids
    assert index.datasets.count(product=ls8_eo3_dataset.product.name, geopolygon=exact3_4326) == 2
    dssids = [ds.id for ds in index.datasets.search(product=ls8_eo3_dataset.product.name, geopolygon=exact3_3577)]
    assert len(dssids) == 2
    assert ls8_eo3_dataset3.id in dssids
    assert ls8_eo3_dataset3.id in dssids
    assert index.datasets.count(product=ls8_eo3_dataset.product.name, geopolygon=exact3_3577) == 2


@pytest.mark.parametrize('datacube_env_name', ('experimental',))
def test_temporal_extents(index,
                          ls8_eo3_dataset, ls8_eo3_dataset2,
                          ls8_eo3_dataset3, ls8_eo3_dataset4):
    start, end = index.products.temporal_extent(ls8_eo3_dataset.product)
    assert start == datetime.datetime(
        2013, 4, 4, 0, 58, 34, 682275,
        tzinfo=datetime.timezone.utc)
    assert end == datetime.datetime(
        2016, 5, 28, 23, 50, 59, 149573,
        tzinfo=datetime.timezone.utc)
    start2, end2 = index.products.temporal_extent(ls8_eo3_dataset.product.name)
    assert start == start2 and end == end2
    start2, end2 = index.datasets.temporal_extent([
        ls8_eo3_dataset.id, ls8_eo3_dataset2.id,
        ls8_eo3_dataset3.id, ls8_eo3_dataset4.id,
    ])
    assert start == start2 and end == end2



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from pathlib import Path
from types import SimpleNamespace

import pytest

from datacube import Datacube
from datacube.testutils import gen_tiff_dataset, mk_test_image

odc_loader = pytest.importorskip("odc.loader")


def test_with_driver(tmpdir):
    tmpdir = Path(str(tmpdir))

    spatial = dict(
        resolution=(15, -15),
        offset=(11230, 1381110),
    )

    nodata = -999
    aa = mk_test_image(96, 64, "int16", nodata=nodata)

    ds, geobox = gen_tiff_dataset(
        [SimpleNamespace(name="aa", values=aa, nodata=nodata)],
        tmpdir,
        prefix="ds1-",
        timestamp="2018-07-19",
        **spatial
    )
    assert ds.time is not None
    mm = ["aa"]
    mm = [ds.product.measurements[k] for k in mm]
    sources = Datacube.group_datasets([ds], "time")

    ds_data = Datacube.load_data(sources, geobox, mm, driver="rio", dask_chunks={})
    assert ds_data is not None
